#!/usr/bin/env python3
r"""
This file contains logic to convert a kwcoco file into an IARPA Site Model.

At a glance the IARPA Site Model is a GeoJSON FeatureCollection with the
following informal schema:

For official documentation about the KWCOCO json format see [kwcoco]_. A formal
json-schema can be found in ``kwcoco.coco_schema``

For official documentation about the IARPA json format see [2, 3]_. A formal
json-schema can be found in ``../../geowatch/rc/site-model.schema.json``.

References:
    .. [kwcoco] https://gitlab.kitware.com/computer-vision/kwcoco
    .. [2] https://infrastructure.smartgitlab.com/docs/pages/api/
    .. [3] https://smartgitlab.com/TE/annotations

SeeAlso:
    * ../tasks/tracking/from_heatmap.py
    * ../tasks/tracking/old_polygon_extraction.py
    * ../tasks/tracking/polygon_extraction.py
    * ../tasks/tracking/utils.py
    * ../../tests/test_tracker.py

Ignore:
    python -m geowatch.cli.run_tracker \
        --in_file /data/joncrall/dvc-repos/smart_expt_dvc/_debug/metrics/bas-fusion/bas_fusion_kwcoco.json \
        --out_site_summaries_fpath /data/joncrall/dvc-repos/smart_expt_dvc/_debug/metrics/bas-fusion/tracking_manifests_bas2/region_models_manifest.json \
        --out_site_summaries_dir /data/joncrall/dvc-repos/smart_expt_dvc/_debug/metrics/bas-fusion/region_models2 \
        --out_kwcoco /data/joncrall/dvc-repos/smart_expt_dvc/_debug/metrics/bas-fusion/bas_fusion_kwcoco_tracked3.json \
        --default_track_fn saliency_heatmaps \
        --append_mode=False \
        --track_kwargs "
            thresh: 0.33
            inner_window_size: 1y
            inner_agg_fn: mean
            norm_ord: 1
            agg_fn: probs
            resolution: 10GSD
            moving_window_size: 1
            min_area_square_meters: 7200
            max_area_square_meters: 9000000
            poly_merge_method: v1
        "

        geowatch visualize /data/joncrall/dvc-repos/smart_expt_dvc/_debug/metrics/bas-fusion/bas_fusion_kwcoco_tracked3.json --smart
"""
import os
import scriptconfig as scfg
import ubelt as ub

if not os.environ.get('_ARGCOMPLETE', ''):
    from geowatch.tasks.tracking import from_heatmap, from_polygon

    _KNOWN_TRACK_FUNCS = {
        'saliency_heatmaps': from_heatmap.TimeAggregatedBAS,
        'saliency_polys': from_polygon.OverlapTrack,
        'class_heatmaps': from_heatmap.TimeAggregatedSC,
        'site_validation': from_heatmap.TimeAggregatedSV,
        'class_polys': from_polygon.OverlapTrack,
        'mono_track': from_polygon.MonoTrack,
    }

    _trackfn_details_docs = ' --- '.join([
        # k + ': ' + ', '.join([field.name for field in v.__dataclass_fields__.values()])
        # if hasattr(v, '__dataclass_fields__') else
        k + ': ' + ', '.join([k for k, v in v.__default__.items()])
        if hasattr(v, '__default__') else
        k + ':?'
        for k, v in _KNOWN_TRACK_FUNCS.items()
    ])
else:
    _trackfn_details_docs = 'na'


try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


USE_NEW_KWCOCO_TRACKS = False


class KWCocoToGeoJSONConfig(scfg.DataConfig):
    """
    Convert KWCOCO to IARPA GeoJSON
    """

    # TODO: can we store metadata in the type annotation?
    # SeeAlso: notes in from_heatmap.__devnote__
    # e.g.
    #
    # in_file : scfg.Coercible(help='Input KWCOCO to convert', position=1) = None
    # in_file : scfg.Type(help='Input KWCOCO to convert', position=1) = None
    # in_file : scfg.PathLike(help='Input KWCOCO to convert', position=1) = None
    #
    # or
    #
    # in_file = None
    # in_file : scfg.Value[help='Input KWCOCO to convert', position=1]

    in_file = scfg.Value(None, required=True, help='Input KWCOCO to convert',
                         position=1, alias=['input_kwcoco'])

    out_kwcoco = scfg.Value(None, help=ub.paragraph(
            '''
            The file path to write the "tracked" kwcoco file to.
            '''))

    out_sites_dir = scfg.Value(None, help=ub.paragraph(
        '''
        The directory where site model geojson files will be written.
        '''), alias=['output_sites_dpath'])

    out_site_summaries_dir = scfg.Value(None, help=ub.paragraph(
        '''
        The directory path where site summary geojson files will be written.
        '''))

    out_sites_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The file path where a manifest of all site models will be written.
        '''), alias=['output_site_manifest_fpath'])

    out_site_summaries_fpath = scfg.Value(None, help=ub.paragraph(
        '''
        The file path where a manifest of all site summary geojson files will
        be written.
        '''))

    in_file_gt = scfg.Value(None, help=ub.paragraph(
            '''
            If available, ground truth KWCOCO to visualize.
            DEPRECATED.
            '''), group='convenience')

    region_id = scfg.Value(None, help=ub.paragraph(
            '''
            ID for region that sites belong to. If None, try to infer
            from kwcoco file.
            '''), group='convenience')

    track_fn = scfg.Value(None, help=ub.paragraph(
            '''
            Function to add tracks. If None, use existing tracks.
            Example:
            'geowatch.tasks.tracking.from_heatmap.TimeAggregatedBAS'
            '''), group='track', mutex_group=1)
    default_track_fn = scfg.Value(None, help=ub.paragraph(
            '''
            String code to pick a sensible track_fn based on the
            contents of in_file. Supported codes are
            ['saliency_heatmaps', 'saliency_polys', 'class_heatmaps',
            'class_polys']. Any other string will be interpreted as the
            image channel to use for 'saliency_heatmaps' (default:
            'salient'). Supported classes are ['Site Preparation',
            'Active Construction', 'Post Construction', 'No Activity'].
            For class_heatmaps, these should be image channels; for
            class_polys, they should be annotation categories.
            '''), group='track', mutex_group=1)
    track_kwargs = scfg.Value('{}', type=str, help=ub.paragraph(
            f'''
            JSON string or path to file containing keyword arguments for
            the chosen TrackFunction. Examples include: coco_dset_gt,
            coco_dset_sc, thresh, key. Any file paths will be loaded as
            CocoDatasets if possible.

            IN THE FUTURE THESE MEMBERS WILL LIKELY BECOME FLAT CONFIG OPTIONS.

            Valid params for each track_fn are: {_trackfn_details_docs}
            '''), group='track')
    viz_out_dir = scfg.Value(None, help=ub.paragraph(
            '''
            Directory to save tracking vizualizations to; if None, don't viz
            '''), group='track')

    site_summary = scfg.Value(None, help=ub.paragraph(
            '''
            A filepath glob or json blob containing either a
            site_summary or a region_model that includes site summaries.
            Each summary found will be added to in_file as
            'Site Boundary' annotations.
            '''), alias=['in_site_summaries'], group='behavior')

    clear_annots = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Clears all annotations before running tracking, so it starts
            from a clean slate.
            TODO: perhaps name to something that takes the value overwrite or
            append?
            '''), group='behavior')
    append_mode = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Append sites to existing region GeoJSON.
            '''), group='behavior')

    boundary_region = scfg.Value(None, help=ub.paragraph(
            '''
            A path or globstring coercable to one or more region files that
            will define the bounds of where the tracker is allowed to predict
            sites. Any site outside of these bounds will be removed.
            '''), group='behavior')
    # filter_out_of_bounds = scfg.Value(False, isflag=True, help=ub.paragraph(
    #         '''
    #         if True, any tracked site outside of the region bounds
    #         (as specified in the site summary) will be removed.
    #         '''), group='behavior')

    sensor_warnings = scfg.Value(True, help='if False, disable sensor warnings')

    time_pad_before = scfg.Value(None, help='A time delta to extend start times')
    time_pad_after = scfg.Value(None, help='A time delta to extend end times')

    #### New Eval 16 params

    smoothing = scfg.Value(None, help=ub.paragraph(
        '''
        if True specify a number between 0 and 1 to smooth the observation
        scores over time.
        '''))

    site_score_thresh = scfg.Value(None, help=ub.paragraph(
        '''
        if specified, then the final a site will be rejected if its site score
        is less than this threshold.
        NOTE: In the future we should separate the steps that assign scores to
        polygons / define polygons and those that postprocess / threshold them
        into another stage. For now we are putting it here.
        '''))


__cli__ = KWCocoToGeoJSONConfig


def _single_geometry(geom):
    import shapely
    return shapely.geometry.shape(geom).buffer(0)


def _ensure_multi(poly):
    """
    Args:
        poly (Union[shapely.geometry.Polygon, shapely.geometry.MultiPolygon])

    Returns:
        shapely.geometry.MultiPolygon
    """
    import shapely
    if isinstance(poly, shapely.geometry.MultiPolygon):
        return poly
    elif isinstance(poly, shapely.geometry.Polygon):
        return shapely.geometry.MultiPolygon([poly])
    else:
        raise TypeError(f'{poly} of type {type(poly)}')


def _combined_geometries(geometry_list):
    import shapely.ops
    return shapely.ops.unary_union(geometry_list).buffer(0)


def _normalize_date(date_str):
    from kwutil import util_time
    return util_time.coerce_datetime(date_str).date().isoformat()


def _join_props(parts):
    # T&E requires a comma and a space
    return ', '.join(parts)


def _split_props(parts):
    # Handle comma only and comma space cases
    return [p.strip() for p in parts.split(',')]


# For MultiPolygon observations. Could be ', '?
# sep = ','


@profile
def coco_create_observation(coco_dset, anns):
    """
    Group kwcoco annotations in the same track (site) and image into one
    Feature in an IARPA site model
    """
    # import geojson
    import kwcoco
    import shapely
    import numpy as np
    from collections import defaultdict
    from geowatch.geoannots import geomodels

    def single_geometry(ann):
        seg_geo = ann['segmentation_geos']
        assert isinstance(seg_geo, dict)
        return _single_geometry(seg_geo)

    @profile
    def _per_image_properties(coco_img: kwcoco.CocoImage):
        """
        Properties defined per-img instead of per-ann, to reduce duplicate
        computation.
        """
        # pick the image that is actually copied to the metrics framework
        # the source field is implied to be a STAC id, but overload it to
        # enable viz during scoring without referring back to the kwcoco file
        # TODO maybe use cache for this instead when STAC id is
        # properly passed through to TA-2?
        source = None
        img = coco_img.img
        bundle_dpath = ub.Path(coco_img.bundle_dpath)
        chan_to_asset = {
            aux['channels']: aux
            for aux in coco_img.iter_asset_objs()
        }
        _candidate_source_chans = {
            'r|g|b', 'rgb', 'pan', 'panchromatic', 'green', 'blue'
        }
        for want_chan in _candidate_source_chans:
            if want_chan in chan_to_asset:
                aux = chan_to_asset[want_chan]
                source = bundle_dpath / aux['file_name']
                break

        # if source is None:
        #     # Fallback to the "primary" filepath (too slow)
        #     source = os.path.abspath(coco_img.primary_image_filepath())

        if source is None:
            try:
                # Note, this will likely fail
                # Pick reasonable source image, we don't have a spec for this
                candidate_keys = [
                    'parent_name', 'parent_file_name', 'name', 'file_name'
                ]
                source = next(filter(None, map(img.get, candidate_keys)))
            except StopIteration:
                raise Exception(f'can\'t determine source of gid {img["gid"]}')

        date = _normalize_date(img['date_captured'])

        return {
            'source': source,
            'observation_date': date,
            'is_occluded': False,  # HACK
            'sensor_name': img['sensor_coarse']
        }

    def single_properties(ann):

        current_phase = coco_dset.cats[ann['category_id']]['name']

        # Give each observation a total score
        score = ann.get('score', 1.0)
        score = max(min(score, 1.0), 0.0)

        # Remember scores for each class
        raw_multi_scores = ann.get('scores', {})

        # Give each obs a status. This lets us decouple kwcoco in existing
        # downstream logic. Might not be needed in the future.
        status = ann.get('status', 'system_confirmed')

        return {
            'type': 'observation',
            'current_phase': current_phase,
            'is_site_boundary': True,  # HACK
            'score': score,
            'cache': {
                'phase_transition_days': ann.get('phase_transition_days', None),
                'raw_multi_scores': raw_multi_scores,
                'status': status
            },
            **image_properties_dct[ann['image_id']]
        }

    def combined_properties(properties_list, geometry_list):
        # list of dicts -> dict of lists for easy indexing
        properties_dcts = {
            k: [dct[k] for dct in properties_list]
            for k in properties_list[0]
        }

        properties = {}

        def _len(geom):
            if isinstance(geom, shapely.geometry.Polygon):
                return 1  # this is probably the case
            elif isinstance(geom, shapely.geometry.MultiPolygon):
                try:
                    return len(geom)
                except TypeError:
                    return len(geom.geoms)
            else:
                raise TypeError(type(geom))

        # per-polygon properties
        for key in ['current_phase', 'is_occluded', 'is_site_boundary']:
            value = []
            for prop, geom in zip(properties_dcts[key], geometry_list):
                value.append(_join_props([str(prop)] * _len(geom)))
            properties[key] = _join_props(value)

        # HACK
        # We are not being scored on multipolygons in SC right now!
        # https://smartgitlab.com/TE/metrics-and-test-framework/-/issues/24
        #
        # When safe, merge class labels for multipolygons so they'll be scored.
        if 0:
            phase = _split_props(properties['current_phase'])
            if len(phase) > 1 and len(set(phase)) == 1:
                properties['current_phase'] = phase[0]

        # identical properties
        for key in ['type', 'source', 'observation_date', 'sensor_name']:
            values = properties_dcts[key]
            assert len(set(values)) == 1
            properties[key] = str(values[0])

        # take area-weighted average score
        weights = np.array([geom.area for geom in geometry_list])
        scores = np.array([float(s) for s in properties_dcts['score']])
        mask = np.isnan(scores)
        scores[mask] = 0
        weights[mask] = 0
        if weights.sum() == 0:
            properties['score'] = 0
        else:
            properties['score'] = np.average(scores, weights=weights)

        # Make cache a list of entries for each polygon.
        # This is for the case of more than one polygon beloning to the
        # observation, which I'm not even sure if that makes sense.
        properties['cache'] = defaultdict(list)
        for cache in properties_dcts['cache']:
            for k, v in cache.items():
                properties['cache'][k].append(v)

        return properties

    image_properties_dct = {}
    gids = {ann['image_id'] for ann in anns}
    for gid in gids:
        coco_img = coco_dset.coco_image(gid).detach()
        image_properties_dct[gid] = _per_image_properties(coco_img)

    geometry_list = list(map(single_geometry, anns))

    properties_list = list(map(single_properties, anns))
    geometry = _ensure_multi(_combined_geometries(geometry_list))
    properties = combined_properties(properties_list, geometry_list)

    observation = geomodels.Observation(geometry=geometry, properties=properties)
    return observation


def predict_phase_changes(site_id, observations):
    """
    Set predicted_phase_transition and predicted_phase_transition_date.

    This should only kick in when the site does not end before the current
    day (latest available image). See tracking.normalize.normalize_phases
    for what happens if the site has ended.

    Args:
        site_id (str): site identifier
        features (List[Dict]): observation feature dictionaries for the site

    Returns:
        dict

    References:
        https://smartgitlab.com/TE/standards/-/wikis/Site-Model-Specification
        https://gitlab.kitware.com/smart/standards-wiki/-/blob/main/Site-Model-Specification.md

    Example:
        >>> from geowatch.geoannots import geomodels
        >>> site = geomodels.SiteModel.random(rng=0, num_observations=20)
        >>> site_id = site.site_id
        >>> observations = list(site.body_features())
        >>> observations[-1]['properties']['cache'] = {'phase_transition_days': [100]}
        >>> predict_phase_changes(site_id, observations)
    """
    import datetime as datetime_mod
    from kwutil import util_time

    feature_properties = [feat['properties'] for feat in observations]

    # Ensure features are in temporal order
    # (they probably are already, but we are being safe)
    feature_properties = sorted(
        feature_properties,
        key=lambda prop: util_time.coerce_datetime(prop['observation_date']))

    feature_phases = [_split_props(prop['current_phase']) for prop in feature_properties]

    def transition_date_from(phase):
        for props, phases in zip(reversed(feature_properties), reversed(feature_phases)):
            if phase in phases:
                _idx = phases.index(phase)
                cache = props.get('cache', {})
                obs_date = util_time.coerce_datetime(props['observation_date'])
                days = cache['phase_transition_days'][_idx]
                pred_delta = datetime_mod.timedelta(days=int(days))
                pred_datetime = (obs_date + pred_delta)
                pred_date = pred_datetime.date()
                return pred_date.isoformat()
        print(f'warning: {site_id=} is missing {phase=}')
        final_date = util_time.coerce_datetime(
            feature_properties[-1]['observation_date'])
        tomorrow = final_date + datetime_mod.timedelta(days=1)
        tomorrow = tomorrow.date()
        return tomorrow.isoformat()

    final_phase = feature_phases[-1]

    if 'Post Construction' in final_phase:
        return {}
    elif 'Active Construction' in final_phase:
        return {
            'predicted_phase_transition': 'Post Construction',
            'predicted_phase_transition_date': transition_date_from('Active Construction')
        }
    elif 'Site Preparation' in final_phase:
        return {
            'predicted_phase_transition': 'Active Construction',
            'predicted_phase_transition_date': transition_date_from('Site Preparation')
        }
    else:
        # raise ValueError(f'missing phases: {site_id=} {all_phases_set=}')
        print(f'missing phases: {site_id=} {final_phase=}')
        return {}


def smooth_observation_scores(observations, smoothing=0.5, smooth_mode='ewma'):
    """
    Add smoothed scores inplace

    Example:
        >>> from geowatch.cli.run_tracker import *  # NOQA
        >>> from geowatch.geoannots import geomodels
        >>> site = geomodels.SiteModel.random(num_observations=15)
        >>> observations = list(site.observations())
        >>> # Add random scores for tests
        >>> import kwarray
        >>> rng = kwarray.ensure_rng()
        >>> for obs in observations:
        >>>     obs['properties']['cache'] = {'raw_multi_scores': [{
        >>>         'No Activity': rng.rand(),
        >>>         'Site Preparation': rng.rand(),
        >>>         'Post Construction': rng.rand(),
        >>>         'Active Construction': rng.rand(),
        >>> }]}
        >>> data1 = [obs['properties']['cache']['raw_multi_scores'][0] for obs in observations]
        >>> smooth_observation_scores(observations, smooth_mode='ewma')
        >>> data2 = [obs['properties']['cache']['smooth_scores'].copy() for obs in observations]
        >>> smooth_observation_scores(observations, smooth_mode='conv3')
        >>> data3 = [obs['properties']['cache']['smooth_scores'].copy() for obs in observations]
        >>> import pandas as pd
        >>> df1 = pd.DataFrame(data1)
        >>> df2 = pd.DataFrame(data2)
        >>> df3 = pd.DataFrame(data3)
        >>> print(df1)
        >>> print(df2)
        >>> print(df3)
    """
    import kwarray
    # Consolidate scores over time. Predict start / end dates.
    obs_multi_scores = [obs['properties']['cache']['raw_multi_scores'] for obs in observations]
    assert all(len(ms) == 1 for ms in obs_multi_scores)
    obs_scores = [ms[0] for ms in obs_multi_scores]
    score_df = kwarray.DataFrameLight.from_dict(obs_scores)
    score_df = score_df.pandas()

    if smooth_mode == 'ewma':
        alpha = (1 - smoothing)
        smoothed = score_df
        smoothed = smoothed.ewm(alpha=alpha, axis=0).mean()
    elif smooth_mode == 'conv3':
        # This didn't do that well in tests, we may be able to remove it.
        import scipy.ndimage
        import numpy as np
        kernel = np.empty(3)
        middle = 2 * (1 - smoothing) / 3 + (1 / 3)
        side = (1 - middle) / 2
        kernel[1] = middle
        kernel[0] = side
        kernel[2] = side
        new_values = scipy.ndimage.convolve1d(score_df.values, kernel, axis=1, mode='constant', cval=0)
        smoothed = score_df.copy()
        smoothed.values[:] = new_values
    else:
        raise KeyError(smooth_mode)

    # HACK: Our score dict mixes scores of different types, but we need to just
    # predict activity phase labels here.
    if 'ac_salient' in smoothed.columns:
        phase_candidates = smoothed.drop(['ac_salient'], axis=1)
    else:
        phase_candidates = smoothed

    new_labels = phase_candidates.idxmax(axis=1).values.tolist()
    # old_labels = [o['properties']['current_phase'] for o in observations]

    new_scores = smoothed.to_dict('records')
    for obs, scores, label in zip(observations, new_scores, new_labels):
        obs['properties']['current_phase'] = label
        obs['properties']['cache']['smooth_scores'] = scores

    if 0:
        from geowatch.tasks.tracking import phase
        new_labels = phase.class_label_smoothing(
            new_labels, transition_probs='v7', emission_probs='v7')

    # phase.REGISTERED_EMMISSION_PROBS['default']
    # phase.REGISTERED_TRANSITION_PROBS['default']
    # phase.viterbi(new_labels, )


def classify_site(site, config):
    """
    Modify a site inplace with classifications.

    Given a site with extracted and scored observations, postprocess the raw
    observation scores and make site-level predictions.
    """
    import numpy as np

    site_header = site.header
    observations = list(site.observations())

    if config.smoothing is not None and config.smoothing > 0:
        smooth_observation_scores(observations, smoothing=config.smoothing)

    ACTIVE_LABELS = {'Site Preparation', 'Active Construction'}

    # Score each observation
    per_obs_scores = []
    for obs in observations:
        obs_cache = obs['properties']['cache']
        if 'smooth_scores' in obs_cache:
            obs_scores = obs_cache['smooth_scores']
        elif 'raw_multi_scores' in obs_cache:
            obs_scores = obs_cache['raw_multi_scores'][0]
        else:
            # fallback to hard classifications
            obs_scores = {obs['properties']['current_phase']: 1.0}

        if 'ac_salient' in obs_scores:
            # Hard code to use "ac_salient" score if available.
            obs_score = obs_scores['ac_salient']
        else:
            active_scores = ub.udict(obs_scores) & ACTIVE_LABELS
            if active_scores:
                obs_score = max(active_scores.values())
            else:
                # final fallback
                obs_score = obs['properties'].get('score', 1.0)
        per_obs_scores.append(obs_score)

    # Site score is the maximum observation score.
    site_score = site_header['properties'].get('score', 1.0)
    status = site_header['properties'].get('status', 'system_confirmed')
    start_date = site_header['properties']['start_date']
    end_date = site_header['properties']['end_date']
    site_header['properties'].setdefault('cache', {})

    if per_obs_scores:
        site_score = max(per_obs_scores)

    FILTER_INACTIVE_SITES = True
    if FILTER_INACTIVE_SITES:
        # HACKS FOR EVAL 15 to get things done quicky.
        curr_labels = [o['properties']['current_phase'] for o in observations]
        flags = [lbl in ACTIVE_LABELS for lbl in curr_labels]
        active_idxs = np.where(flags)[0]
        if len(active_idxs):
            first_idx = active_idxs[0]
            last_idx = active_idxs[-1]
            start_date = observations[first_idx].observation_date.date().isoformat()
            end_date = observations[last_idx].observation_date.date().isoformat()
        else:
            # Check to make sure we are in AC/SC. Otherwise use old logic
            if any(c in {'No Activity', 'Post Construction'} for c in curr_labels):
                # Throw out items with no start / end
                status = 'system_rejected'
                site_header['properties']['cache']['reject_reason'] = 'no_active_observations'

    if config.site_score_thresh is not None:
        # Reject sites if we have a threshold here.
        if site_score < config.site_score_thresh:
            status = 'system_rejected'
            site_header['properties']['cache']['reject_reason'] = 'failed_site_score_thresh'

    # Hack to modify time bounds
    import kwutil
    if config.time_pad_before is not None:
        delta_before = kwutil.util_time.timedelta.coerce(config.time_pad_before)
        start_date = (kwutil.util_time.datetime.coerce(start_date) - delta_before).date().isoformat()
    if config.time_pad_after is not None:
        delta_after = kwutil.util_time.timedelta.coerce(config.time_pad_after)
        end_date = (kwutil.util_time.datetime.coerce(end_date) + delta_after).date().isoformat()

    site_header['properties'].update({
        'score': site_score,
        'status': status,
        'start_date': start_date,
        'end_date': end_date,
    })

    site_id = site.site_id
    pred_transition = predict_phase_changes(site_id, observations)
    site_header['properties'].update(pred_transition)


def coco_create_site_header(region_id, site_id, trackid, observations):
    """
    Feature containing metadata about the site

    Returns:
        geomodels.SiteSummary | geomodels.SiteHeader
    """
    import shapely
    import geowatch
    from geowatch.geoannots import geomodels

    geom_list = [_single_geometry(feat['geometry']) for feat in observations]
    geometry = _combined_geometries(geom_list)

    # site and site_summary features must be polygons
    if isinstance(geometry, shapely.geometry.MultiPolygon):
        if len(geometry.geoms) == 1:
            geometry = geometry.geoms[0]
        else:
            print(f'warning: {region_id=} {site_id=} {trackid=} has multi-part site geometry')
            geometry = geometry.convex_hull

    # these are strings, but sorting should be correct in isoformat

    status_set = [obs['properties']['cache'].get('status', ['system_confirmed'])
                  for obs in observations]
    status_set = set(ub.flatten(status_set))

    # https://smartgitlab.com/TE/annotations/-/wikis/Annotation-Status-Types#for-site-models-generated-by-performersalgorithms
    # system_confirmed, system_rejected, or system_proposed
    # TODO system_proposed pre val-net
    assert len(status_set) == 1, f'inconsistent {status_set=} for {trackid=}'
    status = status_set.pop()

    dates = sorted([obs.observation_date for obs in observations])
    start_date = dates[0].date().isoformat()
    end_date = dates[-1].date().isoformat()

    PERFORMER_ID = 'kit'

    site_header = geomodels.SiteHeader.empty()
    site_header['geometry'] = geometry
    site_header['properties'].update({
        'site_id': site_id,
        'region_id': region_id,

        'version': geowatch.__version__,  # Shouldn't this be a schema version?

        'status': status,
        'model_content': 'proposed',
        'score': 1.0,  # TODO does this matter?

        'start_date': start_date,
        'end_date': end_date,

        'originator': PERFORMER_ID,
        'validated': 'False'
    })
    site_header.infer_mgrs()
    return site_header


@profile
def convert_kwcoco_to_iarpa(coco_dset, default_region_id=None):
    """
    Convert a kwcoco coco_dset to the IARPA JSON format

    Args:
        coco_dset (kwcoco.CocoDataset):
            a coco dataset, but requires images are geotiffs as well as certain
            special fields.

    Returns:
        dict: sites
            dictionary of json-style data in IARPA site format

    Example:
        >>> import geowatch
        >>> from geowatch.cli.run_tracker import *  # NOQA
        >>> from geowatch.tasks.tracking.normalize import run_tracking_pipeline
        >>> from geowatch.tasks.tracking.from_polygon import MonoTrack
        >>> import ubelt as ub
        >>> coco_dset = geowatch.coerce_kwcoco('geowatch-msi', heatmap=True, geodata=True, dates=True)
        >>> coco_dset = run_tracking_pipeline(
        >>>     coco_dset, track_fn=MonoTrack, overwrite=False,
        >>>     sensor_warnings=False)
        >>> videos = coco_dset.videos()
        >>> videos.set('name', ['DM_R{:03d}'.format(vidid) for vidid in videos])
        >>> sites = convert_kwcoco_to_iarpa(coco_dset)
        >>> print(f'{len(sites)} sites')
        >>> if 0:  # validation fails
        >>>     import jsonschema
        >>>     SITE_SCHEMA = geowatch.rc.load_site_model_schema()
        >>>     for site in sites:
        >>>         jsonschema.validate(site, schema=SITE_SCHEMA)
        >>> elif 0:  # but this works if metrics are available
        >>>     import tempfile
        >>>     import json
        >>>     from iarpa_smart_metrics.evaluation import SiteStack
        >>>     for site in sites:
        >>>         with tempfile.NamedTemporaryFile() as f:
        >>>             json.dump(site, open(f.name, 'w'))
        >>>             SiteStack(f.name)
    """
    import itertools as it
    counter = it.count()
    sites = []
    for vidid, video in coco_dset.index.videos.items():
        region_id = video.get('name', default_region_id)
        gids = coco_dset.index.vidid_to_gids[vidid]
        region_annots = coco_dset.images(gids).annots
        region_track_ids = sorted(set(ub.flatten(region_annots.lookup('track_id'))))

        # SUPER HACK:
        # There isn't a good mechanism to get the region id,
        # This is fragile, but currently works. Need better mechanism.
        if '_CLUSTER' in region_id:
            region_id = region_id.split('_CLUSTER')[0]

        for site_idx, trackid in zip(counter, region_track_ids):
            site = coco_track_to_site(coco_dset, trackid, region_id, site_idx)
            sites.append(site)

    return sites


@profile
def coco_track_to_site(coco_dset, trackid, region_id, site_idx=None):
    """
    Turn a kwcoco track into an IARPA site model or site summary
    """
    # import geojson
    from geowatch.geoannots.geomodels import SiteModel
    import geowatch

    # get annotations in this track sorted by frame_index
    annots = coco_dset.annots(track_id=trackid)
    gids, anns = annots.gids, annots.objs

    observations = []
    for gid, _anns in ub.group_items(anns, gids).items():
        feat = coco_create_observation(coco_dset, _anns)
        observations.append(feat)
    # for i in range(len(observations)):
    #     observations[i]['properties'].setdefault('cache', {})
    #     observations[i]['properties']['cache']['trackid'] = trackid

    # HACK to passthrough site_summary IDs
    if geowatch.tasks.tracking.utils.trackid_is_default(trackid):
        if site_idx is None:
            site_idx = trackid
        site_id = '_'.join((region_id, str(site_idx).zfill(4)))
    else:
        # HACK:
        # TODO make more robust
        site_id = trackid
        region_id = '_'.join(site_id.split('_')[:2])

    site_header = coco_create_site_header(region_id, site_id, trackid, observations)
    site = SiteModel([site_header] + observations)

    return site


def _coerce_site_summaries(site_summary_or_region_model,
                              default_region_id=None):
    """
    Possible input formats:
        - file path
        - globbed file paths
        - stringified json blob
    Leading to a:
        - site summary
        - region model containing site summaries

    Args:
        default_region_id: for summaries.
            Region models should already have a region_id.
        strict: if True, raise error on unknown input

    Returns:
        List[Tuple[str, Dict]]
           Each tuple is a (region_id, site_summary) pair
    """
    from kwgis.utils import util_gis
    from geowatch.geoannots import geomodels
    import jsonschema

    TRUST_REGION_SCHEMA = 0

    geojson_infos = list(util_gis.coerce_geojson_datas(
        site_summary_or_region_model, format='json', allow_raw=True))

    # validate the json
    site_summaries = []

    # # debug mode is for comparing against a set of known GT site models
    # DEBUG_MODE = 1
    # if DEBUG_MODE:
    #     SITE_SUMMARY_POS_STATUS = {
    #         'positive_annotated',
    #         'system_proposed', 'system_confirmed',
    #         'positive_annotated_static',  # TODO confirm
    #         'negative',
    #     }
    # else:
    #     # TODO handle positive_partial
    #     SITE_SUMMARY_POS_STATUS = {
    #         'positive_annotated',
    #         'system_proposed', 'system_confirmed',
    #     }

    for info in geojson_infos:
        data = info['data']

        if not isinstance(data, dict):
            raise AssertionError(
                f'unknown site summary {type(data)=}'
            )

        try:  # is this a region model?

            region_model = geomodels.RegionModel(**data)

            if TRUST_REGION_SCHEMA:
                region_model.fixup()
                region_model.validate(strict=False)

            region_model._validate_quick_checks()

            _summaries = [
                f for f in region_model.site_summaries()
                if f['properties']['status'] not in {'system_rejected'}
            ]
            # region_id = region_model.region_id
            # TODO: handle default region-id if needed

            site_summaries.extend(_summaries)

        except jsonschema.ValidationError:
            # In this case we expect the input to be a list of site summaries.
            # However, we really shouldn't hit this case.
            raise AssertionError(
                'Jon thinks we wont hit this case. '
                'If you see this error, he is wrong and the error can be removed. '
                'Otherwise we should remove this extra code')
            site_summary = site_summary_or_region_model
            # site_summary_tups.append((default_region_id, site_summary))
            site_summaries.append(site_summary)

    return site_summaries


def assign_sites_to_videos(coco_dset, site_summaries, viz_out_dir=None):
    """
    Compute assignments between which sites summaries should be projected onto
    which videos for scoring.
    """
    from kwgis.utils import util_gis
    from geowatch.geoannots.geomodels import RegionModel
    import rich
    from geowatch.utils import kwcoco_extensions

    site_idx_to_vidid = []
    unassigned_site_idxs = []

    # Get a GeoDataFrame with geometry for each coco video and each site.
    video_gdf = kwcoco_extensions.covered_video_geo_regions(coco_dset)

    sitesum_model = RegionModel(features=site_summaries)
    sitesum_gdf = sitesum_model.pandas_summaries()

    # Compute which sites overlap which videos
    site_idx_to_video_idx = util_gis.geopandas_pairwise_overlaps(sitesum_gdf, video_gdf)

    # For each site, chose a single video assign to
    assigned_idx_pairs = []
    for site_idx, video_idxs in site_idx_to_video_idx.items():
        if len(video_idxs) == 0:
            unassigned_site_idxs.append(site_idx)
        elif len(video_idxs) == 1:
            assigned_idx_pairs.append((site_idx, video_idxs[0]))
        else:
            qshape = sitesum_gdf.iloc[site_idx]['geometry']
            candidates = video_gdf.iloc[video_idxs]
            overlaps = []
            for dshape in candidates['geometry']:
                iarea = qshape.intersection(dshape).area
                uarea = qshape.area
                iofa = iarea / uarea
                overlaps.append(iofa)
            idx = ub.argmax(overlaps)
            assigned_idx_pairs.append((site_idx, video_idxs[idx]))

    for site_idx, video_idx in assigned_idx_pairs:
        video_id = video_gdf.iloc[video_idx]['video_id']
        site_idx_to_vidid.append((site_idx, video_id))

    assigned_vidids = set([t[1] for t in site_idx_to_vidid])
    n_unassigned_sites = len(unassigned_site_idxs)
    n_assigned_sites = len(site_idx_to_vidid)
    n_total_sites = len(site_summaries)
    n_assigned_vids = len(assigned_vidids)
    n_total_vids = coco_dset.n_videos

    color, punc = '', '.'
    if unassigned_site_idxs:
        color, punc = '[yellow]', '!'
    rich.print(f'{color}There were {n_unassigned_sites} sites that has no video overlaps{punc}')
    rich.print(f'There were {n_assigned_sites} / {n_total_sites} assigned site summaries')
    rich.print(f'There were {n_assigned_vids} / {n_total_vids} assigned videos')

    if viz_out_dir is not None:
        rich.print('Drawing assignments')
        # Draw assignments
        import kwplot
        from geowatch.utils import util_kwplot
        import kwimage
        import numpy as np
        # kwplot.autompl()
        with kwplot.BackendContext(backend='agg'):
            figman = util_kwplot.FigureManager(dpath=viz_out_dir, verbose=1)
            fig = figman.figure(fnum=1, doclf=True)
            ax = fig.gca()
            bounds = kwimage.Boxes(np.concatenate([
                video_gdf.geometry.bounds.values,
                sitesum_gdf.geometry.bounds.values,
            ]), 'ltrb')
            minx, miny, maxx, maxy = bounds.bounding_box().data[0]
            ax.set_xlim(minx, maxx)
            ax.set_ylim(miny, maxy)
            video_gdf.geometry.plot(ax=ax, facecolor='none', edgecolor='black')
            sitesum_gdf.geometry.plot(ax=ax, facecolor='none', edgecolor='blue')
            figman.set_figtitle('Tracking Bounds-to-Cluster Assignment')
            figman.finalize('tracking_bounds_to_cluster_assignment.png')

    return site_idx_to_vidid


def add_site_summary_to_kwcoco(possible_summaries,
                               coco_dset,
                               default_region_id=None,
                               viz_out_dir=None):
    """
    Add a site summary(s) to a kwcoco dataset as a set of polygon annotations.
    These annotations will have category "Site Boundary", 1 track per summary.

    This function is mainly for SC. The "possible_summaries" indicate regions
    flagged by BAS (which could also be truth data if we are evaluating SC
    independently) that need SC processing. We need to associate these and
    place them in the correct videos so we can process those areas.
    """
    import rich
    import geowatch
    from geowatch.utils import kwcoco_extensions
    from kwutil import util_time
    import kwcoco
    # input validation
    print(f'possible_summaries={possible_summaries}')
    print(f'coco_dset={coco_dset}')
    print(f'default_region_id={default_region_id}')

    if default_region_id is None:
        default_region_id = ub.peek(coco_dset.index.name_to_video)

    # TODO: maintain system-rejected sites?
    site_summary_or_region_model = possible_summaries
    site_summaries = _coerce_site_summaries(
        site_summary_or_region_model, default_region_id)
    print(f'found {len(site_summaries)} site summaries')

    site_summary_cid = coco_dset.ensure_category(geowatch.heuristics.SITE_SUMMARY_CNAME)

    print('Searching for assignment between requested site summaries and the kwcoco videos')

    # TODO: Should likely refactor to unify this codepath with reproject
    # annotations.

    # FIXME: need to include a temporal component to this assignment.
    # Also, should probably do this in UTM instead of CRS84

    # Compute Assignment between site summaries / coco videos.
    site_idx_to_vidid = assign_sites_to_videos(coco_dset, site_summaries,
                                               viz_out_dir=viz_out_dir)

    print('warping site boundaries to pxl space...')

    for site_idx, video_id in site_idx_to_vidid:

        site_summary = site_summaries[site_idx]
        site_id = site_summary['properties']['site_id']

        # get relevant images
        images = coco_dset.images(video_id=video_id)

        # and their dates
        image_dates = [util_time.coerce_datetime(d).date()
                       for d in images.lookup('date_captured')]

        assert image_dates == sorted(image_dates), 'images are not in order'
        first_date = image_dates[0]
        last_date = image_dates[-1]

        start_dt = site_summary.start_date
        end_dt = site_summary.end_date
        if start_dt is None:
            if end_dt is not None and end_dt.date() < first_date:
                rich.print(f'[yellow]warning: end_dt before first_date! end_dt={end_dt.date()} first_date={first_date}')
                continue
            start_date = first_date
        else:
            start_date = start_dt.date()
        if end_dt is None:
            if start_date > last_date:
                rich.print(f'[yellow]warning: start_date after last_date! start_date={start_date} last_date={last_date}')
                continue
            end_date = last_date
        else:
            end_date = end_dt.date()

        if end_date < start_date:
            raise AssertionError(ub.codeblock(
                f'''
                warning: end_date before start_date!
                site_summary={site_summary}
                start_date={start_date}
                end_date={start_date}
                ''')
            )

        flags = [start_date <= d <= end_date for d in image_dates]
        images = images.compress(flags)

        if USE_NEW_KWCOCO_TRACKS:
            try:
                track_id = coco_dset.add_track(name=site_id)
            except kwcoco.exceptions.DuplicateAddError:
                rich.print(f'[yellow]warning: site_summary {site_id} already in dset!')
                raise
        else:
            track_id = site_id
            # Seems broken
            if track_id in images.get('track_id', None):
                rich.print(f'[yellow]warning: site_summary {track_id} already in dset!')

        # apply site boundary as polygons
        poly_crs84_geojson = site_summary['geometry']
        for img in images.objs:
            # Add annotations in CRS84 geo-space, we will project to pixel
            # space in a later step
            coco_dset.add_annotation(
                image_id=img['id'],
                category_id=site_summary_cid,
                segmentation_geos=poly_crs84_geojson,
                track_id=track_id
            )

    print('Projecting regions to pixel coords')
    kwcoco_extensions.warp_annot_segmentations_from_geos(coco_dset)
    rich.print(f'[green]Done projecting: {coco_dset.n_annots} annotations')
    return coco_dset


@profile
def main(argv=None, **kwargs):
    """
    Example:
        >>> # test BAS and default (SC) modes
        >>> from geowatch.cli.run_tracker import *  # NOQA
        >>> from geowatch.cli.run_tracker import main
        >>> from geowatch.demo import smart_kwcoco_demodata
        >>> from kwgis.utils import util_gis
        >>> import json
        >>> import kwcoco
        >>> import ubelt as ub
        >>> # run BAS on demodata in a new place
        >>> import geowatch
        >>> coco_dset = geowatch.coerce_kwcoco('geowatch-msi', heatmap=True, geodata=True, dates=True)
        >>> dpath = ub.Path.appdir('geowatch', 'test', 'tracking', 'main0').ensuredir()
        >>> coco_dset.reroot(absolute=True)
        >>> coco_dset.fpath = dpath / 'bas_input.kwcoco.json'
        >>> coco_dset.clear_annotations()
        >>> coco_dset.dump(coco_dset.fpath, indent=2)
        >>> region_id = 'dummy_region'
        >>> regions_dir = dpath / 'regions/'
        >>> bas_coco_fpath = dpath / 'bas_output.kwcoco.json'
        >>> sc_coco_fpath = dpath / 'sc_output.kwcoco.json'
        >>> bas_fpath = dpath / 'bas_sites.json'
        >>> sc_fpath = dpath / 'sc_sites.json'
        >>> # Run BAS
        >>> argv = bas_args = [
        >>>     '--in_file', coco_dset.fpath,
        >>>     '--out_site_summaries_dir', str(regions_dir),
        >>>     '--out_site_summaries_fpath',  str(bas_fpath),
        >>>     '--out_kwcoco', str(bas_coco_fpath),
        >>>     '--track_fn', 'saliency_heatmaps',
        >>>     '--sensor_warnings', 'False',
        >>>     '--track_kwargs', json.dumps({
        >>>        'thresh': 1e-9, 'min_area_square_meters': None,
        >>>        'max_area_square_meters': None,
        >>>        'polygon_simplify_tolerance': 1}),
        >>> ]
        >>> main(argv)
        >>> # Run SC on the same dset, but with BAS pred sites removed
        >>> sites_dir = dpath / 'sites'
        >>> argv = sc_args = [
        >>>     '--in_file', coco_dset.fpath,
        >>>     '--out_sites_dir', str(sites_dir),
        >>>     '--out_sites_fpath', str(sc_fpath),
        >>>     '--out_kwcoco', str(sc_coco_fpath),
        >>>     '--track_fn', 'class_heatmaps',
        >>>     '--site_summary', str(bas_fpath),
        >>>     '--sensor_warnings', 'False',
        >>>     '--track_kwargs', json.dumps(
        >>>         {'thresh': 1e-9, 'min_area_square_meters': None, 'max_area_square_meters': None,
        >>>          'polygon_simplify_tolerance': 1, 'key': 'salient'}),
        >>> ]
        >>> main(argv)
        >>> # Check expected results
        >>> bas_coco_dset = kwcoco.CocoDataset(bas_coco_fpath)
        >>> sc_coco_dset = kwcoco.CocoDataset(sc_coco_fpath)
        >>> bas_trackids = bas_coco_dset.annots().lookup('track_id', None)
        >>> sc_trackids = sc_coco_dset.annots().lookup('track_id', None)
        >>> print('bas_trackids = {}'.format(ub.urepr(bas_trackids, nl=1)))
        >>> print('sc_trackids = {}'.format(ub.urepr(sc_trackids, nl=1)))
        >>> assert len(bas_trackids) and None not in bas_trackids
        >>> assert len(sc_trackids) and None not in sc_trackids
        >>> summaries = list(util_gis.coerce_geojson_datas(bas_fpath, format='dataframe'))
        >>> sites = list(util_gis.coerce_geojson_datas(sc_fpath, format='dataframe'))
        >>> import pandas as pd
        >>> sc_df = pd.concat([d['data'] for d in sites])
        >>> bas_df = pd.concat([d['data'] for d in summaries])
        >>> ssum_rows = bas_df[bas_df['type'] == 'site_summary']
        >>> site_rows = sc_df[sc_df['type'] == 'site']
        >>> obs_rows = sc_df[sc_df['type'] == 'observation']
        >>> assert len(site_rows) > 0
        >>> assert len(ssum_rows) > 0
        >>> assert len(ssum_rows) == len(site_rows)
        >>> assert len(obs_rows) > len(site_rows)
        >>> # Cleanup
        >>> #dpath.delete()

    Example:
        >>> # test resolution
        >>> from geowatch.cli.run_tracker import *  # NOQA
        >>> from geowatch.cli.run_tracker import main
        >>> import geowatch
        >>> dset = geowatch.coerce_kwcoco('geowatch-msi', heatmap=True, geodata=True, dates=True)
        >>> dpath = ub.Path.appdir('geowatch', 'test', 'tracking', 'main1').ensuredir()
        >>> out_fpath = dpath / 'resolution_test.kwcoco.json'
        >>> regions_dir = dpath / 'regions'
        >>> bas_fpath = dpath / 'bas_sites.json'
        >>> import json
        >>> track_kwargs = json.dumps({
        >>>         'resolution': '10GSD',
        >>>         'min_area_square_meters': 1000000,  # high area threshold filters results
        >>>         'max_area_square_meters': None,
        >>>         'thresh': 1e-9,
        >>> })
        >>> kwargs = {
        >>>     'in_file': str(dset.fpath),
        >>>     'out_site_summaries_dir': str(regions_dir),
        >>>     'out_site_summaries_fpath':  str(bas_fpath),
        >>>     'out_kwcoco': str(out_fpath),
        >>>     'track_fn': 'saliency_heatmaps',
        >>>     'track_kwargs': track_kwargs,
        >>>     'sensor_warnings': False,
        >>> }
        >>> argv = []
        >>> # Test case for no results
        >>> main(argv=argv, **kwargs)
        >>> from kwgis.utils import util_gis
        >>> assert len(list(util_gis.coerce_geojson_datas(bas_fpath))) == 0
        >>> # Try to get results here
        >>> track_kwargs = json.dumps({
        >>>         'resolution': '10GSD',
        >>>         'min_area_square_meters': None,
        >>>         'max_area_square_meters': None,
        >>>         'thresh': 1e-9,
        >>> })
        >>> kwargs = {
        >>>     'in_file': str(dset.fpath),
        >>>     'out_site_summaries_dir': str(regions_dir),
        >>>     'out_site_summaries_fpath':  str(bas_fpath),
        >>>     'out_kwcoco': str(out_fpath),
        >>>     'track_fn': 'saliency_heatmaps',
        >>>     'track_kwargs': track_kwargs,
        >>>     'sensor_warnings': False,
        >>> }
        >>> argv = []
        >>> main(argv=argv, **kwargs)
        >>> assert len(list(util_gis.coerce_geojson_datas(bas_fpath))) > 0

    Example:
        >>> # xdoctest: +REQUIRES(--slow)
        >>> # test a more complicated track function
        >>> import geowatch
        >>> from geowatch.cli.run_tracker import demo
        >>> import kwcoco
        >>> import geowatch
        >>> import ubelt as ub
        >>> # make a new BAS dataset
        >>> coco_dset = geowatch.coerce_kwcoco('geowatch-msi', heatmap=True, geodata=True)
        >>> #coco_dset.images().set('sensor_coarse', 'S2')
        >>> for img in coco_dset.imgs.values():
        >>>     img['sensor_coarse'] = 'S2'
        >>> coco_dset.remove_categories(coco_dset.cats.keys())
        >>> coco_dset.fpath = 'bas.kwcoco.json'
        >>> # TODO make serializable, check set() and main()
        >>> coco_dset.dump(coco_dset.fpath, indent=2)
        >>> # make a new SC dataset
        >>> coco_dset_sc = smart_kwcoco_demodata.demo_kwcoco_with_heatmaps(
        >>>     num_videos=2)
        >>> for img in coco_dset_sc.imgs.values():
        >>>     img['sensor_coarse'] = 'S2'
        >>> coco_dset_sc.remove_categories(coco_dset_sc.cats.keys())
        >>> for img in coco_dset_sc.imgs.values():
        >>>     for aux, key in zip(img['auxiliary'],
        >>>                         ['Site Preparation', 'Active Construction',
        >>>                          'Post Construction', 'No Activity']):
        >>>         aux['channels'] = key
        >>> coco_dset_sc.fpath = 'sc.kwcoco.json'
        >>> coco_dset_sc.dump(coco_dset_sc.fpath, indent=2)
        >>> regions_dir = 'regions/'
        >>> sites_dir = 'sites/'
        >>> # moved this to a separate function for length
        >>> demo(coco_dset, regions_dir, coco_dset_sc, sites_dir, cleanup=True)
    """
    cmdline = True
    if isinstance(argv, list) and argv == []:
        # Workaround an issue in scriptconfig tha twill be fixed in 0.7.8
        # after that, remove cmdline completely
        cmdline = False

    try:
        args = KWCocoToGeoJSONConfig.cli(cmdline=cmdline, argv=argv, data=kwargs, strict=True)
    except Exception:
        print('Error when parsing CLI arguments')
        print(f'kwargs = {ub.urepr(kwargs, nl=1)}')
        print(f'argv = {ub.urepr(argv, nl=1)}')
        print(f'cmdline = {ub.urepr(cmdline, nl=1)}')
        raise
    import rich
    rich.print('args = {}'.format(ub.urepr(args, nl=1)))

    import geojson
    import json
    import kwcoco
    import pandas as pd
    import safer
    import geowatch

    from kwcoco.util import util_json
    from kwutil.util_yaml import Yaml
    from geowatch.geoannots import geomodels
    from geowatch.utils import process_context
    from kwgis.utils import util_gis

    coco_fpath = ub.Path(args.in_file)

    if args.out_sites_dir is not None:
        args.out_sites_dir = ub.Path(args.out_sites_dir)

    if args.out_site_summaries_dir is not None:
        args.out_site_summaries_dir = ub.Path(args.out_site_summaries_dir)

    if args.out_sites_fpath is not None:
        if args.out_sites_dir is None:
            raise ValueError(
                'The directory to store individual sites must be specified')
        args.out_sites_fpath = ub.Path(args.out_sites_fpath)
        if not str(args.out_sites_fpath).endswith('.json'):
            raise ValueError('out_sites_fpath should have a .json extension')

    if args.out_site_summaries_fpath is not None:
        if args.out_site_summaries_dir is None:
            raise ValueError(
                'The directory to store individual site summaries must be '
                'specified')
        args.out_site_summaries_fpath = ub.Path(args.out_site_summaries_fpath)
        if not str(args.out_site_summaries_fpath).endswith('.json'):
            raise ValueError('out_site_summaries_fpath should have a .json extension')

    # load the track kwargs
    track_kwargs = Yaml.coerce(args.track_kwargs)
    assert isinstance(track_kwargs, dict)

    # Read the kwcoco file
    coco_dset = kwcoco.CocoDataset.coerce(coco_fpath)
    # HACK read auxiliary kwcoco files
    # these aren't used for core tracking algos, supported for legacy
    if args.in_file_gt is not None:
        gt_dset = kwcoco.CocoDataset.coerce(args.in_file_gt)
    else:
        gt_dset = None
    for k, v in track_kwargs.items():
        if isinstance(v, str) and os.path.isfile(v):
            try:
                track_kwargs[k] = kwcoco.CocoDataset.coerce(v)
            except json.JSONDecodeError:  # TODO make this smarter
                pass

    if args.clear_annots:
        coco_dset.clear_annotations()

    pred_info = coco_dset.dataset.get('info', [])

    tracking_output = {
        'type': 'tracking_result',
        'info': [],
        'files': [],
    }
    # Args will be serailized in kwcoco, so make sure it can be coerced to json
    jsonified_config = util_json.ensure_json_serializable(args.asdict())
    walker = ub.IndexableWalker(jsonified_config)
    for problem in util_json.find_json_unserializable(jsonified_config):
        bad_data = problem['data']
        walker[problem['loc']] = str(bad_data)

    # TODO: ensure all args are resolved here.
    info = tracking_output['info']

    PROCESSCONTEXT_INFO_HACK = True
    proc_context_kwargs = {}
    if PROCESSCONTEXT_INFO_HACK:
        # TODO: now that we are no longer doing subsets and unions in the tracker
        # we dont need to maintain the old "info" here. In fact if we do, we get
        # circular references. In order to maintain compatability with older output
        # we are going to clear the old info and continue to use extra pred info,
        # but in the future after we are sure nothing depends on this, we should
        # simply remove the extra kwarg, as it will be implicitly maintained.
        proc_context_kwargs['extra'] = {'pred_info': pred_info}
        # Clear out the old info to prevent circular references while we do the
        # above hack. Remove the next line once we fix the above hack.
        coco_dset.dataset['info'] = []

    proc_context = process_context.ProcessContext(
        name='geowatch.cli.run_tracker', type='process',
        config=jsonified_config,
        track_emissions=False,
        **proc_context_kwargs,
    )

    proc_context.start()
    info.append(proc_context.obj)

    # Pick a track_fn
    # HACK remove potentially conflicting annotations as well
    # we shouldn't have saliency annots when we want class or vice versa
    CLEAN_DSET = 1
    class_cats = [cat['name'] for cat in geowatch.heuristics.CATEGORIES]
    saliency_cats = ['salient']

    track_fn = args.track_fn
    if track_fn is None:
        track_fn = (
            geowatch.tasks.tracking.abstract_classes.NoOpTrackFunction
            if args.default_track_fn is None else
            args.default_track_fn
        )

    if isinstance(track_fn, str):
        # TODO: we should be able to let the user know about these algorithms
        # and parameters. Can jsonargparse help here?
        if CLEAN_DSET:
            if 'class_' in track_fn:
                coco_dset.remove_categories(saliency_cats)
            else:
                coco_dset.remove_categories(class_cats)

        track_fn = _KNOWN_TRACK_FUNCS.get(track_fn, None)
        if track_fn is None:
            raise RuntimeError('Old code would have evaled track_fn, we dont want to do that. '
                               'Please change your code to specify a known track function')

    if track_fn is None:
        raise KeyError(
            f'Unknown Default Track Function: {args.default_track_fn} not in {list(_KNOWN_TRACK_FUNCS.keys())}')

    if args.boundary_region is not None:
        region_infos = list(util_gis.coerce_geojson_datas(
            args.boundary_region, format='json', allow_raw=True,
            desc='load boundary regions'))
        region_parts = []
        for info in region_infos:
            # Need to deterimine which one to use
            region_model = geomodels.RegionModel(**info['data'])
            region_gdf = region_model.pandas_region()
            region_parts.append(region_gdf)
        boundary_regions_gdf = pd.concat(region_parts).reset_index()
        video_gdf = coco_video_gdf(coco_dset)
        video_region_assignments = assign_videos_to_regions(video_gdf, boundary_regions_gdf)
    else:
        if args.site_summary is None:
            rich.print('[yellow]Warning: No boundary regions or site summaries specified')
        else:
            rich.print('No boundary regions specified')
        boundary_regions_gdf = None
        video_region_assignments = None

    # add site summaries (site boundary annotations)
    if args.site_summary is not None:

        coco_dset = add_site_summary_to_kwcoco(args.site_summary, coco_dset,
                                               args.region_id,
                                               viz_out_dir=args.viz_out_dir)
        if 0:
            # Going to test to see if removing this helps anything
            cid = coco_dset.name_to_cat[geowatch.heuristics.SITE_SUMMARY_CNAME]['id']
            coco_dset = coco_dset.subset(coco_dset.index.cid_to_gids[cid])
            print('restricting dset to videos with site_summary annots: ',
                  set(coco_dset.index.name_to_video))
            assert coco_dset.n_images > 0, 'no valid videos!'
    else:
        # Note: this may be a warning if we use this in a pipeline where the
        # inputed kwcoco files already have polygons reprojected (and in that
        # case if there are legidimately no sites to score!) But in our main
        # use case where site_summary should be specified, this is an error so
        # we are treating it that way for now.
        if track_kwargs.get('boundaries_as') == 'polys' and not coco_dset.n_annots:
            raise Exception(ub.codeblock(
                '''
                You requested scoring boundaries as polygons, but the dataset
                does not have any annotations and no site summaries were
                specified to add them.  You probably forgot to specify the site
                summaries that should be scored!
                '''))

    # TODO: in the case that we are NOT reestimating polygon boundaries, then
    # we should pass through any previous system rejected sites.

    print(f'track_fn={track_fn}')
    """
    ../tasks/tracking/normalize.py
    """
    coco_dset = geowatch.tasks.tracking.normalize.run_tracking_pipeline(
        coco_dset, track_fn=track_fn, gt_dset=gt_dset,
        viz_out_dir=args.viz_out_dir, sensor_warnings=args.sensor_warnings,
        **track_kwargs)

    if boundary_regions_gdf is not None:
        print('Cropping to boundary regions')
        coco_remove_out_of_bound_tracks(coco_dset, video_region_assignments)

    # Measure how long tracking takes
    proc_context.stop()

    rich.print('[green] Finished main tracking phase')

    out_kwcoco = args.out_kwcoco

    if out_kwcoco is not None:
        coco_dset = coco_dset.reroot(absolute=True, check=False)
        # Add tracking audit data to the kwcoco file
        coco_info = coco_dset.dataset.get('info', [])
        coco_info.append(proc_context.obj)
        coco_dset.fpath = out_kwcoco
        ub.Path(out_kwcoco).parent.ensuredir()
        print(f'write to coco_dset.fpath={coco_dset.fpath}')
        coco_dset.dump(out_kwcoco, indent=2)

    verbose = 1

    # Convert scored kwcoco tracks to sites models
    all_sites = convert_kwcoco_to_iarpa(
        coco_dset, default_region_id=args.region_id)
    print(f'{len(all_sites)=}')

    # Postprocess / classify sites
    config = args

    # for site in all_sites:
    #     site.header['properties'].setdefault('cache', {})
    #     print(site.header['properties']['status'], site.header['properties']['score'], site.header['properties']['cache'].get('reject_reason'))

    for site in ub.ProgIter(all_sites, desc='classify sites'):
        classify_site(site, config)

    # for site in all_sites:
    #     site.header['properties'].setdefault('cache', {})
    #     print(site.header['properties']['status'], site.header['properties']['score'], site.header['properties']['cache'].get('reject_reason'))

    if args.out_sites_dir is not None:
        # write sites to disk
        sites_dir = ub.Path(args.out_sites_dir).ensuredir()
        site_fpaths = []
        for site in ub.ProgIter(all_sites, desc='writing sites', verbose=verbose):
            site_props = site['features'][0]['properties']
            assert site_props['type'] == 'site'
            site_fpath = sites_dir / (site_props['site_id'] + '.geojson')
            site_fpaths.append(os.fspath(site_fpath))

            with safer.open(site_fpath, 'w', temp_file=not ub.WIN32) as f:
                geojson.dump(site, f, indent=2)

    output_dirs = ub.oset()
    if args.out_sites_fpath is not None:
        site_tracking_output = tracking_output.copy()
        site_tracking_output['files'] = site_fpaths
        out_sites_fpath = ub.Path(args.out_sites_fpath)
        out_sites_fpath.parent.ensuredir()
        output_dirs.add(out_sites_fpath.parent)
        print(f'Write tracked site result to {out_sites_fpath}')
        with safer.open(out_sites_fpath, 'w', temp_file=not ub.WIN32) as file:
            json.dump(site_tracking_output, file, indent='    ')

    # Convert site models to site summaries
    if args.out_site_summaries_dir is not None:

        site_summary_dir = ub.Path(args.out_site_summaries_dir).ensuredir()
        output_dirs.add(site_summary_dir)
        # write site summaries to region models on disk
        groups = ub.group_items(all_sites, lambda site: site.header['properties']['region_id'])

        site_summary_fpaths = []
        for region_id, sites in groups.items():

            # Converting sites to a region model makes them site summaries
            sites = geomodels.SiteModelCollection(sites)
            new_summaries = sites.as_region_model()
            region_fpath = site_summary_dir / (region_id + '.geojson')

            if args.append_mode and region_fpath.is_file():
                new_region = geomodels.RegionModel.coerce(region_fpath)
                new_region.features.extend(list(new_summaries.site_summaries()))
                if verbose:
                    print(f'writing to existing region {region_fpath}')
            else:
                new_region = new_summaries
                if verbose:
                    print(f'writing to new region {region_fpath}')

            site_summary_fpaths.append(os.fspath(region_fpath))
            with safer.open(region_fpath, 'w', temp_file=not ub.WIN32) as f:
                geojson.dump(new_region, f, indent=2)

    if args.out_site_summaries_fpath is not None:
        site_summary_tracking_output = tracking_output.copy()
        site_summary_tracking_output['files'] = site_summary_fpaths
        out_site_summaries_fpath = ub.Path(args.out_site_summaries_fpath)
        out_site_summaries_fpath.parent.ensuredir()
        output_dirs.add(out_site_summaries_fpath.parent)
        print(f'Write tracked site summary result to {out_site_summaries_fpath}')
        with safer.open(out_site_summaries_fpath, 'w', temp_file=not ub.WIN32) as file:
            json.dump(site_summary_tracking_output, file, indent='    ')

    for dpath in output_dirs:
        rich.print(f'Tracked Result Dir: [link={dpath}]{dpath}[/link]')

    if args.viz_out_dir is not None:
        rich.print(f'Tracking Viz: [link={args.viz_out_dir}]{args.viz_out_dir}[/link]')


def coco_video_gdf(coco_dset):
    # TODO: rectify with covered_video_geo_regions
    import pandas as pd
    from kwgis.utils import util_gis
    from geowatch.geoannots.geococo_objects import CocoGeoVideo
    crs84 = util_gis.get_crs84()
    crs84_parts = []
    for video in coco_dset.videos().objs:
        coco_video = CocoGeoVideo(video=video, dset=coco_dset)
        utm_part = coco_video.wld_corners_gdf
        crs84_part = utm_part.to_crs(crs84)
        crs84_parts.append(crs84_part)
    video_gdf = pd.concat(crs84_parts).reset_index()
    return video_gdf


def assign_videos_to_regions(video_gdf, boundary_regions_gdf):
    """
    Assign each video to a region (usually for BAS)
    """
    from kwgis.utils import util_gis
    idx1_to_idxs2 = util_gis.geopandas_pairwise_overlaps(video_gdf, boundary_regions_gdf)
    video_region_assignments = []
    for idx1, idxs2 in idx1_to_idxs2.items():
        video_row = video_gdf.iloc[idx1]
        video_name = video_row['name']
        region_rows = boundary_regions_gdf.iloc[idxs2]

        if len(idxs2) == 0:
            raise AssertionError(ub.paragraph(
                f'''
                The coco video {video_name} has no intersecting candidate
                boundary regions in the {len(boundary_regions_gdf)} that were
                searched.
                '''))
        elif len(idxs2) > 1:
            # In the case that there are multiple candidates, use the one with
            # the highest IOU and emit a warning.
            isect_area = region_rows.intersection(video_row['geometry']).area
            union_area = region_rows.union(video_row['geometry']).area
            iou = isect_area / union_area
            iou = iou.sort_values(ascending=False)
            region_rows = region_rows.loc[iou.index]
            region_name = region_rows.iloc[0]['region_id']
            region_geom = region_rows.iloc[0]['geometry']

            import warnings
            warnings.warn(ub.paragraph(
                f'''
                The coco video {video_name} has {len(region_rows)} intersecting
                candidate boundary regions. Choosing region {region_name} with
                the highest IOU of {iou.iloc[0]:0.4f}.

                The runner up regions are:
                {ub.urepr(region_rows.iloc[1:3]['region_id'].tolist(), nl=0)}
                with ious:
                {ub.urepr(iou.iloc[1:3].tolist(), precision=4, nl=0)}
                '''))
        else:
            region_name = region_rows.iloc[0]['region_id']
            region_geom = region_rows.iloc[0]['geometry']

        video_region_assignments.append((video_name, region_name, region_geom))
    return video_region_assignments


def coco_remove_out_of_bound_tracks(coco_dset, video_region_assignments):
    # Remove any tracks that are outside of region bounds.
    # First find which regions correspond to which videos.
    from kwgis.utils import util_gis
    from shapely.geometry import shape
    import rich
    import geopandas as gpd
    crs84 = util_gis.get_crs84()

    # Actually remove the offending annots
    to_remove_trackids = set()
    all_track_ids = set()
    for assign in video_region_assignments:
        video_name, region_name, region_geom = assign
        video_id = coco_dset.index.name_to_video[video_name]['id']
        video_imgs = coco_dset.images(video_id=video_id)
        video_aids = list(ub.flatten(video_imgs.annots))
        video_annots = coco_dset.annots(video_aids)
        all_track_ids.update(video_annots.lookup('track_id'))
        annot_geos = gpd.GeoDataFrame([
            {
                'annot_id': obj['id'],
                'geometry': shape(obj['segmentation_geos']),
            } for obj in video_annots.objs],
            columns=['annot_id', 'geometry'], crs=crs84)

        is_oob = annot_geos.intersection(region_geom).is_empty
        inbound_annots = video_annots.compress(~is_oob)
        outofbounds_annots = video_annots.compress(is_oob)

        # Only remove tracks that are always out of bounds.
        inbound_tracks = set(inbound_annots.lookup('track_id'))
        outofbound_tracks = set(outofbounds_annots.lookup('track_id'))
        always_outofbound_tracks = outofbound_tracks - inbound_tracks
        to_remove_trackids.update(always_outofbound_tracks)

    to_remove_aids = []
    for tid in to_remove_trackids:
        to_remove_aids.extend(list(coco_dset.annots(track_id=tid)))

    if to_remove_aids:
        num_remove_tracks = len(to_remove_trackids)
        num_remove_annots = len(to_remove_aids)
        num_total_tracks = len(all_track_ids)
        num_total_annots = coco_dset.n_annots
        rich.print(ub.paragraph(
            f'''
            [yellow]Removing {num_remove_tracks} / {num_total_tracks} out-of-bounds
            tracks with {num_remove_annots} / {num_total_annots} annotations
            '''))
    else:
        rich.print('[green]All annotations are in bounds')

    coco_dset.remove_annotations(to_remove_aids)


def demo(coco_dset, regions_dir, coco_dset_sc, sites_dir, cleanup=True):
    import json
    from tempfile import NamedTemporaryFile
    bas_args = [
        coco_dset.fpath,
        '--out_site_summaries_dir',
        regions_dir,
        '--track_fn',
        'geowatch.tasks.tracking.from_heatmap.TimeAggregatedBAS',
    ]
    # run BAS on it
    main(bas_args)
    # reload it with tracks
    # coco_dset = kwcoco.CocoDataset(coco_dset.fpath)
    # run SC on both of them
    sc_args = [
        '--out_site_sites_dir',
        sites_dir,
        '--track_fn',
        'geowatch.tasks.tracking.from_heatmap.TimeAggregatedSC',
    ]
    for vid_name, vid in coco_dset_sc.index.name_to_video.items():
        gids = coco_dset_sc.index.vidid_to_gids[vid['id']]
        sub_dset = coco_dset_sc.subset(gids)
        tmpfile = NamedTemporaryFile()
        sub_dset.fpath = tmpfile.name
        sub_dset.dump(sub_dset.fpath)
        region = json.load(
            open(os.path.join(regions_dir, f'{vid_name}.geojson')))
        for site in [
                f for f in region['features']
                if f['properties']['type'] == 'site_summary'
        ]:
            print('running site ' + site['properties']['site_id'])
            main([
                sub_dset.fpath, '--track_kwargs',
                '{"boundaries_as": "none"}'
            ] + sc_args)
            # '--site_summary', json.dumps(site)])
    if cleanup:
        for pth in os.listdir(regions_dir):
            os.remove(os.path.join(regions_dir, pth))
        os.removedirs(regions_dir)
        for pth in os.listdir(sites_dir):
            os.remove(os.path.join(sites_dir, pth))
        os.removedirs(sites_dir)
        if not os.path.isabs(coco_dset.fpath):
            os.remove(coco_dset.fpath)
        if not os.path.isabs(coco_dset_sc.fpath):
            os.remove(coco_dset_sc.fpath)


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the kMaX-DeepLab meta architecture."""
import functools
from typing import Any, Dict, Text

import tensorflow as tf
from absl import logging
from deeplab2 import common
from deeplab2 import config_pb2
from deeplab2.data import dataset
from deeplab2.model import builder
from deeplab2.model import utils
from deeplab2.model.decoder import max_deeplab
from deeplab2.model.post_processor import post_processor_builder


class KMaXDeepLab(tf.keras.Model):
    """This class represents the kMaX-DeepLab meta architecture."""

    def __init__(self,
                 config: config_pb2.ExperimentOptions,
                 dataset_descriptor: dataset.DatasetDescriptor):
        """Initializes a kMaX-DeepLab architecture.

        Args:
          config: A config_pb2.ExperimentOptions configuration.
          dataset_descriptor: A dataset.DatasetDescriptor.

        Raises:
          ValueError: If normalization type is not one of ['sync_batchnorm',
            'batchnorm', 'layernorm'].
        """
        super(KMaXDeepLab, self).__init__(name='kMaXDeepLab')

        # We hard code self._auxiliary_output_number for now.
        self._auxiliary_output_number = 6

        if config.trainer_options.solver_options.use_sync_batchnorm:
            logging.info('Synchronized Batchnorm is used.')
            norm_layer = functools.partial(
                tf.keras.layers.experimental.SyncBatchNormalization,
                momentum=config.trainer_options.solver_options.batchnorm_momentum,
                epsilon=config.trainer_options.solver_options.batchnorm_epsilon)
        else:
            logging.info('Standard (unsynchronized) Batchnorm is used.')
            norm_layer = functools.partial(
                tf.keras.layers.BatchNormalization,
                momentum=config.trainer_options.solver_options.batchnorm_momentum,
                epsilon=config.trainer_options.solver_options.batchnorm_epsilon)

        # The auxiliary predictor does not include the auxiliary semantic head.
        def auxiliary_predictor_func():
            return max_deeplab.MaXDeepLab(
                config.model_options.decoder,
                config.model_options.max_deeplab,
                ignore_label=dataset_descriptor.ignore_label,
                bn_layer=norm_layer,
                use_auxiliary_semantic_decoder=False,
                use_auxiliary_semantic_head=False,
                activation='gelu')

        pretrained_weights_path = None
        if config.model_options.backbone.HasField('pretrained_weights'):
            pretrained_weights_path = config.model_options.backbone.pretrained_weights
        self._pixel_encoder = builder.create_kmax_meta_pixel_encoder(
            pixel_encoder_name=self._infer_pixel_encoder_name_from_config(config),
            drop_path_keep_prob=config.model_options.backbone.drop_path_keep_prob,
            input_shape=list(config.train_dataset_options.crop_size) + [3],
            norm_layer=norm_layer,
            pretrained_weights_path=pretrained_weights_path)

        # Currently, some properties (e.g., high_resolution_output_stride,
        # num_mask_slots) of kMaX are hard-coded for simplicity.
        if 'cityscapes' in dataset_descriptor.dataset_name:
            high_resolution_output_stride = 2
            num_mask_slots = 256
            use_auxiliary_semantic_decoder = False
            transformer_decoder_drop_path_keep_prob = 0.8
        elif 'coco' in dataset_descriptor.dataset_name:
            high_resolution_output_stride = 4
            num_mask_slots = 128
            use_auxiliary_semantic_decoder = True
            transformer_decoder_drop_path_keep_prob = 0.8
        elif 'ade20k' in dataset_descriptor.dataset_name:
            high_resolution_output_stride = 4
            num_mask_slots = 128
            use_auxiliary_semantic_decoder = True
            transformer_decoder_drop_path_keep_prob = 1.0
        else:
            logging.info('kMaX configs for COCO is used for untested datasets.')
            high_resolution_output_stride = 4
            num_mask_slots = 128
            use_auxiliary_semantic_decoder = True
            transformer_decoder_drop_path_keep_prob = 0.8

        self._pixel_decoder = builder.create_kmax_meta_pixel_decoder(
            norm_layer=norm_layer,
            high_resolution_output_stride=high_resolution_output_stride)

        self._transformer_decoder = builder.create_kmax_meta_transformer_decoder(
            norm_layer=norm_layer,
            num_mask_slots=num_mask_slots,
            transformer_decoder_drop_path_keep_prob=(
                transformer_decoder_drop_path_keep_prob),
            auxiliary_predictor_func=auxiliary_predictor_func)

        # Notably, decoder/max_deeplab.MaXDeepLab is in fact the MaX-DeepLab
        # prediction head instead of its decoder.
        self._predictor = max_deeplab.MaXDeepLab(
            config.model_options.decoder,
            config.model_options.max_deeplab,
            ignore_label=dataset_descriptor.ignore_label,
            bn_layer=norm_layer,
            use_auxiliary_semantic_decoder=use_auxiliary_semantic_decoder,
            use_auxiliary_semantic_head=True,
            activation='gelu')

        self._post_processor = post_processor_builder.get_post_processor(
            config, dataset_descriptor)

    def _infer_pixel_encoder_name_from_config(self, config):
        return config.model_options.backbone.name.replace('kmax_', '')

    def _forward_step(self, input_tensor, training=False):
        encoder_pixel_feature = self._pixel_encoder(input_tensor, training=training)
        decoder_pixel_feature = self._pixel_decoder(
            encoder_pixel_feature, training=training)
        decoder_pixel_and_mask_feature = self._transformer_decoder(
            decoder_pixel_feature, training=training)
        pred_dict = self._predictor(decoder_pixel_and_mask_feature,
                                    training=training)
        return pred_dict

    def _inference(self,
                   input_tensor: tf.Tensor,
                   training: bool = False,
                   post_processing: bool = True) -> Dict[Text, Any]:
        """Performs an (optionally multi-scale) inference pass."""
        pred_dict = self._forward_step(input_tensor, training=training)
        result_dict = {}
        for output_type, output_value in pred_dict.items():
            # We remove auxiliary outputs from final result_dict during inference.
            if output_type not in [common.PRED_AUXILIARY_OUTPUTS]:
                result_dict[output_type] = output_value
        # Post-process the results.
        if post_processing:
            result_dict.update(self._post_processor(result_dict))
        return result_dict

    def call(self,
             input_tensor: tf.Tensor,
             training: bool = False,
             post_processing: bool = True) -> Dict[Text, Any]:
        """Performs a forward pass.

        Args:
          input_tensor: An input tensor of type tf.Tensor with shape [batch, height,
            width, channels] of [batch, num_frames, heigh, width, channels]. The
            input tensor should contain batches of RGB images.
          training: A boolean flag indicating whether training behavior should be
            used (default: False).
          post_processing: A boolean flag indicating whether post processing
            (self._post_processor) should be used during testing (default: True).

        Returns:
          A dictionary containing the results of the specified DeepLab architecture.
          The results are bilinearly upsampled to input size before returning.
        """
        # Normalize the input in the same way as Inception. We normalize it outside
        # the encoder so that we can extend encoders to different backbones without
        # copying the normalization to each encoder. We normalize it after data
        # preprocessing because it is faster on TPUs than on host CPUs. The
        # normalization should not increase TPU memory consumption because it does
        # not require gradient.
        input_tensor = input_tensor / 127.5 - 1.0
        if training:
            result_dict = self._forward_step(input_tensor, training=training)
            result_dict = self._prepare_auxiliary_results(result_dict)
        else:
            result_dict = self._inference(input_tensor, training, post_processing)
        return result_dict

    @property
    def auxiliary_output_number(self) -> int:
        return self._auxiliary_output_number

    @property
    def checkpoint_items(self) -> Dict[Text, Any]:
        items = dict(encoder=self._pixel_encoder,
                     pixel_decoder=self._pixel_decoder,
                     transformer_decoder=self._transformer_decoder)
        items.update(self._predictor.checkpoint_items)
        return items

    def _prepare_auxiliary_results(self, result_dict):
        """Prepares and resizes auxiliary predictions to the target size.

        This function prepares auxiliary outputs for loss computation. It also
        resizes the PRED_PIXEL_SPACE_MASK_LOGITS_KEY and
        PRED_PIXEL_SPACE_NORMALIZED_FEATURE_KEY in the result_dict to the target
        height and width.

        Args:
          result_dict: A dictionary storing prediction results to be resized.

        Returns:
          Prepared auxiliary_result_list.
        """
        target_h, target_w = (
            result_dict[
                common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY].get_shape().as_list()[1:3])
        auxiliary_result_list = list(result_dict[common.PRED_AUXILIARY_OUTPUTS])
        for auxiliary_idx in range(len(auxiliary_result_list)):
            auxiliary_result_dict = auxiliary_result_list[auxiliary_idx]
            for key, value in auxiliary_result_dict.items():
                if key in [common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY,
                           common.PRED_PIXEL_SPACE_NORMALIZED_FEATURE_KEY]:
                    auxiliary_result_dict[key] = utils.resize_bilinear(
                        value, [target_h, target_w])
            auxiliary_result_list[auxiliary_idx] = auxiliary_result_dict
        result_dict[common.PRED_AUXILIARY_OUTPUTS] = tuple(auxiliary_result_list)
        return result_dict



from __future__ import annotations

from typing import Tuple, Any
from collections.abc import Callable

import torch
from torch import nn

from .base_node import Node, NodeContainer
from ..utils import is_generator_empty


class TensorNode(Node):
    '''Subclass of node specialzed for nodes that
    stores tensor (subclass of torch.Tensor called RecorderTensor)
    '''
    def __init__(
        self,
        tensor: torch.Tensor,
        depth: int,
        parents: NodeContainer[Node] | Node | None = None,
        children: NodeContainer[Node] | Node | None = None,
        name: str = 'tensor',
        context: Any | None = None,
        is_aux: bool = False,
        main_node: TensorNode | None = None,
        parent_hierarchy: dict[int, ModuleNode | FunctionNode] | None = None,
    ):

        super(TensorNode, self).__init__(
            depth, parents, children, name,
        )
        self.tensor_id = id(tensor)
        self.tensor_shape = tuple(tensor.shape)
        self.name = name
        self.is_aux = is_aux
        self.main_node = self if main_node is None else main_node
        self.context = [] if context is None else context
        self.parent_hierarchy = {} if parent_hierarchy is None else parent_hierarchy
        self.set_node_id()

    def set_node_id(self, children_id: int | str | None = None) -> None:
        if children_id is None:
            self.node_id = (
                f'{id(self.main_node)}' if self.is_aux and self.main_node
                else f'{id(self)}'
            )
        else:
            self.node_id = f'{id(self)}-{children_id}'


class ModuleNode(Node):
    '''Subclass of node specialzed for storing torch Module info
    '''
    def __init__(
        self,
        module_unit: nn.Module,
        depth: int,
        parents: NodeContainer[Node] | Node | None = None,
        children: NodeContainer[Node] | Node | None = None,
        name: str = 'module-node',
        output_nodes: NodeContainer[Node] | None = None,
    ) -> None:
        super(ModuleNode, self).__init__(
            depth, parents, children, name
        )
        self.compute_unit_id = id(module_unit)
        self.is_activation = is_generator_empty(module_unit.parameters())
        self.is_container = not any(module_unit.children())
        self.input_shape: list[Tuple[int, ...]] = []
        self.output_shape: list[Tuple[int, ...]] = []
        self.output_nodes = NodeContainer() if output_nodes is None else output_nodes
        self.set_node_id()

    def set_input_shape(self, input_shape: list[Tuple[int, ...]]) -> None:
        self.input_shape = input_shape

    def set_output_shape(self, output_shape: list[Tuple[int, ...]]) -> None:
        self.output_shape = output_shape

    def add_output_nodes(self, output_node: Node) -> None:
        self.output_nodes.add(output_node)

    def set_node_id(self, output_id: int | str | None = None) -> None:
        '''Sets the id of ModuleNode.
        If no output is given, it sets to value unique to node.
        If output id is given, there are 2 cases:
            1. Parameterless module: id is determined by output_id and id of nn.Module
            2. Module with parameter: id is determined by only id of nn.module object
        This is crucial when rolling recursive modules by identifying them with this id
        mechanism'''
        if output_id is None:
            self.node_id = f'{id(self)}'
        else:
            if self.is_activation:
                # zero-parameter module -> module for activation function, e.g. ReLU
                self.node_id = f'{self.compute_unit_id}-{output_id}'
            else:
                self.node_id = f'{self.compute_unit_id}-'


class FunctionNode(Node):
    '''Subclass of node specialized for nodes
    that does computation (e.g. torch.functions)
    '''
    def __init__(
        self,
        function_unit: Callable[..., Any],
        depth: int,
        parents: NodeContainer[Node] | Node | None = None,
        children: NodeContainer[Node] | Node | None = None,
        name: str = 'function-node',
    ) -> None:
        super(FunctionNode, self).__init__(
            depth, parents, children, name
        )
        self.compute_unit_id = id(function_unit)
        self.is_container = True
        self.input_shape: list[Tuple[int, ...]] = []
        self.output_shape: list[Tuple[int, ...]] = []
        self.set_node_id()
        self.output_nodes = self.children

    def set_input_shape(self, input_shape: list[Tuple[int, ...]]) -> None:
        self.input_shape = input_shape

    def set_output_shape(self, output_shape: list[Tuple[int, ...]]) -> None:
        self.output_shape = output_shape

    def add_output_nodes(self, output_node: Node) -> None:
        self.output_nodes.add(output_node)

    def set_node_id(self, output_id: int | str | None = None) -> None:
        '''Sets the id of FunctionNode.
        If no output is given, it sets to value unique to node.
        If output id is given, id is determined by only id of nn.module object
        This is crucial when rolling recursive modules by identifying them with this id
        mechanism'''
        if output_id is None:
            self.node_id = f'{id(self)}'
        else:
            self.node_id = f'{self.compute_unit_id}-{output_id}'



from collections import OrderedDict
from torch import nn
from torch.nn import functional as F
import torch
import torch.utils.checkpoint

# from pytorch_memlab import profile, profile_every


def humanbytes(B):
    'Return the given bytes as a human friendly KB, MB, GB, or TB string'
    B = float(B)
    KB = float(1024)
    MB = float(KB ** 2)  # 1,048,576
    GB = float(KB ** 3)  # 1,073,741,824
    TB = float(KB ** 4)  # 1,099,511,627,776

    if B < KB:
        return '{0} {1}'.format(B, 'Bytes' if 0 == B > 1 else 'Byte')
    elif KB <= B < MB:
        return '{0:.2f} KB'.format(B / KB)
    elif MB <= B < GB:
        return '{0:.2f} MB'.format(B / MB)
    elif GB <= B < TB:
        return '{0:.2f} GB'.format(B / GB)
    elif TB <= B:
        return '{0:.2f} TB'.format(B / TB)


def get_preact_conv(in_channels, out_channels, kernel_size=3, padding=1, dropout_2d=0.2):
    block = nn.Sequential(
        nn.BatchNorm2d(in_channels),
        nn.ReLU(inplace=True),
        nn.Conv2d(in_channels, out_channels, kernel_size, padding=padding),
        nn.Dropout2d(dropout_2d)
    )
    return block


def _dense_layer_function_factory(norm, relu, conv):
    def bn_function(*inputs):
        concated_features = torch.cat(inputs, 1)
        bottleneck_output = conv(relu(norm(concated_features)))
        return bottleneck_output

    return bn_function


class DenseLayer(nn.Module):
    def __init__(self, in_channels, out_channels, dropout_2d=0.2, efficient=False):
        super(DenseLayer, self).__init__()
        self.add_module('norm', nn.BatchNorm2d(in_channels)),
        self.add_module('relu', nn.ReLU(inplace=True)),
        self.add_module('conv', nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)),
        self.dropout_2d = dropout_2d
        self.efficient = efficient

    def forward(self, *prev_features):
        dense_layer_function = _dense_layer_function_factory(self.norm, self.relu, self.conv)
        if self.efficient and any(prev_feature.requires_grad for prev_feature in prev_features):
            new_features = torch.utils.checkpoint.checkpoint(dense_layer_function, *prev_features)
        else:
            new_features = dense_layer_function(*prev_features)
        if 0 < self.dropout_2d:
            new_features = F.dropout2d(new_features, p=self.dropout_2d, training=self.training)
        return new_features


class SELayer(nn.Module):
    def __init__(self, in_channels, ratio):
        super(SELayer, self).__init__()
        self.block = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),  # Global average pooling
            nn.Flatten(),  # Prepare for fully-connected layers
            nn.Linear(in_channels, in_channels // ratio),
            nn.ReLU(inplace=True),
            nn.Linear(in_channels // ratio, in_channels),
            nn.Sigmoid()
        )

    def forward(self, x):
        excitation = self.block(x)
        x *= excitation[:, :, None, None]
        return x


class DenseBlock(nn.Module):
    def __init__(self, in_channels, n_layers, growth_rate, dropout_2d, return_only_new=False, efficient=False):
        super(DenseBlock, self).__init__()
        assert 0 < n_layers, "n_layers should be at least 1"
        self.in_channels = in_channels
        self.return_only_new = return_only_new

        channels = in_channels
        self.layers = torch.nn.ModuleList()
        for j in range(n_layers):
            # Compute new feature maps
            layer = DenseLayer(channels, growth_rate, dropout_2d=dropout_2d, efficient=efficient)
            self.layers.append(layer)
            channels += growth_rate

        if return_only_new:
            se_layer_in_channel = channels - in_channels  # Remove input, only keep new features
        else:
            se_layer_in_channel = channels
        self.se_layer = SELayer(se_layer_in_channel, ratio=1)

    # @profile_every(1)
    def forward(self, x):
        features = [x]
        for layer in self.layers:
            new_features = layer(*features)
            features.append(new_features)

        if self.return_only_new:
            features = features[1:]

        features = torch.cat(features, 1)
        features = self.se_layer(features)

        return features


def get_transition_down(in_channels, out_channels, dropout_2d=0.2):
    block = nn.Sequential(
        get_preact_conv(in_channels, out_channels, kernel_size=1, padding=0, dropout_2d=dropout_2d),
        nn.MaxPool2d(kernel_size=2, stride=2)
    )
    return block


def cat_non_matching(x1, x2):
    diffY = x1.size()[2] - x2.size()[2]
    diffX = x1.size()[3] - x2.size()[3]

    x2 = F.pad(x2, (diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2))

    # for padding issues, see
    # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
    # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd

    x = torch.cat([x1, x2], dim=1)
    return x


class TransitionUp(nn.Module):
    def __init__(self, in_channels, n_filters_keep):
        super(TransitionUp, self).__init__()
        self.conv_transpose_2d = nn.ConvTranspose2d(in_channels, n_filters_keep, kernel_size=4, stride=2, padding=1)

    def forward(self, x, skip_connection):
        x = self.conv_transpose_2d(x)
        x = cat_non_matching(x, skip_connection)
        return x


class ICTNetBackbone(nn.Module):
    """
    ICTNet model: https://theictlab.org/lp/2019ICTNet.
    """

    def __init__(self, preset_model='FC-DenseNet56', in_channels=3, out_channels=2, n_filters_first_conv=48, n_pool=5, growth_rate=12, n_layers_per_block=4, dropout_2d=0.2, efficient=False):
        super().__init__()

        # --- Handle args
        if preset_model == 'FC-DenseNet56':
            n_pool = 5
            growth_rate = 12
            n_layers_per_block = 4
        elif preset_model == 'FC-DenseNet67':
            n_pool = 5
            growth_rate = 16
            n_layers_per_block = 5
        elif preset_model == 'FC-DenseNet103':
            n_pool = 5
            growth_rate = 16
            n_layers_per_block = [4, 5, 7, 10, 12, 15, 12, 10, 7, 5, 4]
        else:
            n_pool = n_pool
            growth_rate = growth_rate
            n_layers_per_block = n_layers_per_block

        if isinstance(n_layers_per_block, list):
            assert (len(n_layers_per_block) == 2 * n_pool + 1)
        elif isinstance(n_layers_per_block, int):
            n_layers_per_block = [n_layers_per_block] * (2 * n_pool + 1)
        else:
            raise ValueError

        # --- Instantiate layers
        self.first_conv = nn.Conv2d(in_channels, n_filters_first_conv, 3, padding=1)

        # Downsampling path
        channels = n_filters_first_conv
        self.down_dense_blocks = torch.nn.ModuleList()
        self.transition_downs = torch.nn.ModuleList()
        skip_connection_channels = []
        for i in range(n_pool):
            # Dense Block
            self.down_dense_blocks.append(DenseBlock(in_channels=channels, n_layers=n_layers_per_block[i], growth_rate=growth_rate, dropout_2d=dropout_2d, return_only_new=False, efficient=efficient))
            channels += growth_rate * n_layers_per_block[i]
            skip_connection_channels.append(channels)
            # Transition Down
            self.transition_downs.append(get_transition_down(in_channels=channels, out_channels=channels, dropout_2d=dropout_2d))

        # Bottleneck Dense Block
        self.bottleneck_dense_block = DenseBlock(in_channels=channels, n_layers=n_layers_per_block[n_pool], growth_rate=growth_rate, dropout_2d=dropout_2d, return_only_new=True, efficient=efficient)
        up_in_channels = n_layers_per_block[n_pool] * growth_rate  # We will only upsample the new feature maps

        # Upsampling path
        self.transition_ups = torch.nn.ModuleList()
        self.up_dense_blocks = torch.nn.ModuleList()
        for i in range(n_pool):
            # Transition Up (Upsampling + concatenation with the skip connection)
            n_filters_keep = growth_rate * n_layers_per_block[n_pool + i]
            self.transition_ups.append(TransitionUp(in_channels=up_in_channels, n_filters_keep=n_filters_keep))
            up_out_channels = skip_connection_channels[n_pool - i - 1] + n_filters_keep  # After concatenation

            # Dense Block
            # We will only upsample the new feature maps
            self.up_dense_blocks.append(
                DenseBlock(in_channels=up_out_channels, n_layers=n_layers_per_block[n_pool + i + 1], growth_rate=growth_rate,
                           dropout_2d=dropout_2d, return_only_new=True, efficient=efficient))
            up_in_channels = growth_rate * n_layers_per_block[n_pool + i + 1]  # We will only upsample the new feature maps

        # Last layer
        self.final_conv = nn.Conv2d(up_in_channels, out_channels, 1, padding=0)

    # @profile
    def forward(self, x):
        stack = self.first_conv(x)

        skip_connection_list = []
        # print(humanbytes(torch.cuda.memory_allocated()))
        for down_dense_block, transition_down in zip(self.down_dense_blocks, self.transition_downs):
            # Dense Block
            stack = down_dense_block(stack)

            # At the end of the dense block, the current stack is stored in the skip_connections list
            skip_connection_list.append(stack)

            # Transition Down
            stack = transition_down(stack)
        # print(humanbytes(torch.cuda.memory_allocated()))

        skip_connection_list = skip_connection_list[::-1]

        # Bottleneck Dense Block
        # We will only upsample the new feature maps
        stack = self.bottleneck_dense_block(stack)

        # Upsampling path
        # print(humanbytes(torch.cuda.memory_allocated()))
        for transition_up, up_dense_block, skip_connection in zip(self.transition_ups, self.up_dense_blocks, skip_connection_list):
            # Transition Up ( Upsampling + concatenation with the skip connection)
            stack = transition_up(stack, skip_connection)

            # Dense Block
            # We will only upsample the new feature maps
            stack = up_dense_block(stack)
        # print(humanbytes(torch.cuda.memory_allocated()))

        # Final conv
        stack = self.final_conv(stack)

        result = OrderedDict()
        result["out"] = stack

        # print(humanbytes(torch.cuda.memory_allocated()))

        return result


def count_trainable_params(model):
    count = 0
    for param in model.parameters():
        if param.requires_grad:
            count += param.numel()
    return count


def main():
    device = "cuda"
    b = 2
    c = 3
    h = 512
    w = 512
    features = 32

    # Init input
    x = torch.rand((b, c, h, w), device=device)
    print("x: ", x.shape, x.min().item(), x.max().item())

    # # Test SELayer
    # print("--- Test SELayer:")
    # se_layer = SELayer(in_channels=c, ratio=1)
    # y = se_layer(x)
    # print("y: ", y.shape)
    # print("------")

    # # Test DenseBlock
    # print("--- Test DenseBlock:")
    # dense_block = DenseBlock(in_channels=c, n_layers=5, growth_rate=16, dropout_2d=0.2, path_type="down")
    # y, new_y = dense_block(x)
    # print("y: ", y.shape)
    # print("new_y: ", new_y.shape)
    # print("------")

    # # Test transition_down
    # print("--- Test transition_down:")
    # transition_down = get_transition_down(in_channels=c, out_channels=features, dropout_2d=0.2)
    # x_down = transition_down(x)
    # print("x_down: ", x_down.shape)
    # print("------")
    #
    # # Test TransitionUp
    # print("--- Test TransitionUp:")
    # transition_up = TransitionUp(in_channels=features, n_filters_keep=features//2)
    # y = transition_up(x_down, x)
    # print("y: ", y.shape)
    # print("------")

    # Test ICTNetBackboneICTNetBackboneTest SELayer:")
    backbone = ICTNetBackbone(out_channels=features, preset_model="FC-DenseNet103", dropout_2d=0.0, efficient=True)
    print("ICTNetBackbone has {} trainable params".format(count_trainable_params(backbone)))
    # print(backbone)
    backbone.to(device)
    result = backbone(x)
    y = result["out"]
    print("y: ", y.shape)
    print("------")

    print("Back-prop:")
    loss = torch.sum(y)
    loss.backward()


if __name__ == "__main__":
    main()



import ubelt as ub
import os


class Detectron2WrapperBase:

    def __init__(self, config):
        """
        Args:
            config (DetectronFitCLI): the fit CLI config
        """
        self.config = config
        self.dataset_infos = None
        self.cfg = None  # the detectron level config
        self.output_dpath = None  # Where we will actually write to
        self.trainer = None

    def register_datasets(self):
        dataset_paths = {
            'vali': self.config.vali_fpath,
            'train': self.config.train_fpath,
        }
        self.dataset_infos = {}
        with ub.Timer('Registering COCO files with detectron'):
            for key, fpath in dataset_paths.items():
                if fpath is not None:
                    row = register_detectron_dataset(fpath)
                    self.dataset_infos[key] = row
                else:
                    self.dataset_infos[key] = None

    def resolve_cfg(self):
        raise NotImplementedError('abstract')


def register_detectron_dataset(fpath):
    from detectron2.data.datasets import register_coco_instances
    fpath = ub.Path(fpath)
    assert fpath.exists()
    row = {'path': fpath}
    # row['name'] = fpath.name.split('.', 1)[0]
    row['name'] = os.fspath(fpath)
    row['categories'] = parse_json_section(fpath, 'categories')
    register_coco_instances(row['name'], {}, row['path'], row['path'].parent)
    return row


def parse_json_section(fpath, tablename):
    """
    Given a path to a json or compressed json file, attempt to read just one
    specific section. If this section is at the start of the file then this
    operation will be very fast.

    Ignore:
        from geowatch.tasks.detectron2.fit import *  # NOQA
        import kwcoco
        dset = kwcoco.CocoDataset.demo('vidshapes8')
        fpath = dset.fpath
        tablename = 'categories'
        table = parse_json_section(fpath, tablename)
        print(f'table = {ub.urepr(table, nl=1)}')
    """
    from kwcoco.util import ijson_ext
    import zipfile
    if zipfile.is_zipfile(fpath):
        # We have a compressed json file, but we can still read the header
        # fairly quickly.
        zfile = zipfile.ZipFile(fpath)
        names = zfile.namelist()
        assert len(names) == 1
        member = names[0]
        # Stream the header directly from the zipfile.
        file = zfile.open(member, 'r')
    else:
        # Normal json file
        file = open(fpath, 'r')

    with file:
        # import ijson
        # We only expect there to be one info section
        # try:
        #     # Try our extension if the main library fails (due to NaN)
        #     table_iter = ijson.items(file, prefix=tablename)
        #     table = next(info_section_iter)
        # except ijson.IncompleteJSONError:
        # Try our extension if the main library fails (due to NaN)
        # file.seek(0)
        # Nans are too frequent, only use our extension
        table_iter = ijson_ext.items(file, prefix=tablename)
        table = next(table_iter)
    return table



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library for rematerialization.

Incubates a version of tf.recompute_grad that is XLA compatible.

This file is based on the recompute_grad.py in the bigbird codebase [1]:
https://github.com/google-research/bigbird/blob/db06498ec8804c6438111938d8654b66ddaccd5d/bigbird/core/recompute_grad.py

[1] Big Bird: Transformers for Longer Sequences, NeurIPS 2020.
      Manzil Zaheer, Guru Guruganesh, Avinava Dubey, Joshua Ainslie, Chris
      Alberti, Santiago Ontanon, Philip Pham, Anirudh Ravula, Qifan Wang, Li
      Yang, Amr Ahmed.
"""
import collections
import os
import threading
from typing import Deque, List, NamedTuple, Optional, Sequence

import tensorflow.compat.v2 as tf
from absl import logging
# pylint: disable=g-direct-tensorflow-import
from tensorflow.python.framework import ops
from tensorflow.python.ops import custom_gradient


# Remove when https://github.com/tensorflow/tensorflow/pull/45298
# gets merged
def get_variable_by_name(var_name):
    """Retrieves tf.Variable from name in MirroredStrategy (multi-gpu)."""

    # Get all variables, but it will have copies from different replicas
    all_global_vars = ops.get_collection(ops.GraphKeys.GLOBAL_VARIABLES)

    def _replica_filter(var):
        """Filter out variables from different context."""
        try:
            return var_name == var.op.name
        except AttributeError:
            return False

    candidate_vars = list(filter(_replica_filter, all_global_vars))

    if len(candidate_vars) >= 1:
        # Filter out non-trainable variables.
        candidate_vars = [v for v in candidate_vars if v.trainable]
    else:
        raise ValueError('Unsuccessful at finding variable {}.'.format(var_name))

    if len(candidate_vars) == 1:
        return candidate_vars[0]
    elif len(candidate_vars) > 1:
        raise ValueError(
            'Unsuccessful at finding trainable variable {}. '
            'Number of candidates: {}. '
            'Candidates: {}'.format(var_name, len(candidate_vars), candidate_vars))
    else:
        # The variable is not trainable.
        return None


custom_gradient.get_variable_by_name = get_variable_by_name


class RecomputeContext(
    NamedTuple('RecomputeContext', [
        ('is_recomputing', bool),
        ('seed', tf.Tensor),
        ('children', Deque['RecomputeContext']), ])):
    """Context for recomputation.

    Attributes:
      is_recomputing: Whether we are in a recomputation phase.
      seed: Scalar integer tensor that should be used with stateless random ops
        for deterministic behavior and correct computation of the gradient.
      children: Nested `RecomputeContext` instances. Used internally by
        `recompute_grad` to track nested instances of `RecomputeContext`.
    """

    def __enter__(self):
        return _context_stack.push(self)

    def __exit__(self, exc_type, exc_value, traceback):
        _context_stack.pop(self)


# Simplified version of `_DefaultStack` in
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/framework/ops.py.
class _ContextStack(threading.local):
    """A thread-local stack for providing implicit recompute contexts."""

    def __init__(self):
        super(_ContextStack, self).__init__()
        self._stack = []

    def top(self) -> Optional[RecomputeContext]:
        return self._stack[-1] if self._stack else None

    def push(self, context: RecomputeContext):
        self._stack.append(context)
        return context

    def pop(self, context: RecomputeContext):
        if self._stack[-1] is not context:
            raise AssertionError('Nesting violated for RecomputeContext.')
        self._stack.pop()


_context_stack = _ContextStack()


def get_recompute_context() -> Optional[RecomputeContext]:
    """Returns the current recomputing context if it exists."""
    return _context_stack.top()


# Adapted from
# https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/ops/control_flow_util.py.
def _get_containing_xla_context(graph: tf.Graph) -> Optional[object]:
    """Returns the first ancestor `XLAControlFlowContext` in the `graph`."""
    ctxt = graph._get_control_flow_context()  # pylint: disable=protected-access
    while ctxt:
        if ctxt.IsXLAContext():
            return ctxt
        ctxt = ctxt.outer_context
    return None


def _in_xla_context(graph: Optional[tf.Graph] = None) -> bool:
    """Detects whether we are in an XLA context."""
    if '--tf_xla_auto_jit=2' in os.environ.get('TF_XLA_FLAGS', ''):
        return True
    graph = tf.compat.v1.get_default_graph() if graph is None else graph
    while True:
        if _get_containing_xla_context(graph) is not None:
            return True
        try:
            graph = graph.outer_graph
        except AttributeError:
            return False


def _force_data_dependency(
        first_compute: Sequence[tf.Tensor],
        then_compute: Sequence[tf.Tensor]) -> List[tf.Tensor]:
    """Forces all of `then_compute` to depend on all of `first_compute`.

    Uses a dummy data dependency, which is useful when running on TPUs because
    XLA ignores control dependencies. Only supports float arguments.

    Args:
      first_compute: Sequence of `Tensor`s to be executed before `then_compute`.
      then_compute: Sequence of `Tensor`s to executed after `first_compute`.

    Returns:
      Sequence of `Tensor`s with same length of `then_compute`.

    Raises:
      ValueError: if ranks are unknown or types are not floating.
    """

    def _first_element(x):
        if x.shape.ndims is None:
            raise ValueError('Rank of Tensor %s must be known' % x)
        ndims = x.shape.ndims
        begin = tf.zeros(ndims, dtype=tf.int32)
        size = tf.ones(ndims, dtype=tf.int32)
        return tf.reshape(tf.slice(x, begin, size), [])

    first_compute_sum = tf.add_n(
        [_first_element(x) for x in first_compute if x is not None])
    dtype = first_compute_sum.dtype
    if not dtype.is_floating:
        raise ValueError('_force_data_dependency only supports floating dtypes.')
    zero = tf.cast(0.0, first_compute_sum.dtype) * first_compute_sum
    then_compute_sequence = [
        x + tf.cast(zero, x.dtype) if x is not None else None
        for x in tf.nest.flatten(then_compute)
    ]
    return tf.nest.pack_sequence_as(then_compute, then_compute_sequence)


def _make_seed_if_none(seed: Optional[tf.Tensor]) -> tf.Tensor:
    """Uses the global generator to make a seed if necessary."""
    if seed is not None:
        return seed
    generator = tf.random.experimental.get_global_generator()
    # The two seeds for stateless random ops don't have individual semantics and
    # are scrambled together, so providing one seed is fine. This makes it easier
    # for users to provide a local seed without worrying about integer overflow.
    # See `make_seeds` in
    # https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/ops/stateful_random_ops.py.
    try:
        return generator.uniform_full_int([], tf.int32, name='recompute_grad_seed')
    except (RuntimeError, TypeError, ValueError, tf.errors.NotFoundError) as e:
        # For a number of reasons, the above operation can fail like using multiple
        # graphs or toggling between eager and graph modes. Reset the generator.
        logging.warn('Resetting the generator. %s: %s', type(e), e)
        tf.random.experimental.set_global_generator(None)
        generator = tf.random.experimental.get_global_generator()
        return generator.uniform_full_int([], tf.int32, name='recompute_grad_seed')


def recompute_grad(f, seed=None):
    """An eager-compatible version of recompute_grad.

    For f(*args, **kwargs), this supports gradients with respect to args, or to
    gradients with respect to any variables residing in the kwarg 'variables'.
    Note that for keras layer and model objects, this is handled automatically.

    Warning: If `f` was originally a tf.keras Model or Layer object, `g` will not
    be able to access the member variables of that object, because `g` returns
    through the wrapper function `inner`. When recomputing gradients through
    objects that inherit from keras, we suggest keeping a reference to the
    underlying object around for the purpose of accessing these variables.

    Args:
      f: function `f(*x)` that returns a `Tensor` or sequence of `Tensor` outputs.
      seed: Optional seed for random ops. `seed` should an integer scalar
        `Tensor`. When compiling to XLA, `seed` must have dtype `tf.int32`. If
        `seed` is not provided one will be generated.

    Returns:
     A function `g` that wraps `f`, but which recomputes `f` on the backwards
     pass of a gradient call.
    """

    @tf.custom_gradient
    def inner(*args, **kwargs):
        """Inner function closure for calculating gradients."""
        # Detect when we're nested and in the backwards pass, so we don't generate
        # an additional seed.
        parent_context = get_recompute_context()
        if parent_context is not None and parent_context.is_recomputing:
            # Use the cached context in the recomputation phase.
            with parent_context.children.popleft()._replace(
                    is_recomputing=True) as context:
                result = f(*args, **kwargs)
        else:
            with RecomputeContext(
                    is_recomputing=False,
                    seed=_make_seed_if_none(seed),
                    children=collections.deque()) as context:
                result = f(*args, **kwargs)
                # In the forward pass, build up a tree of recomputation contexts.
                if parent_context is not None and not parent_context.is_recomputing:
                    parent_context.children.append(context)

        def grad(*dresult, **grad_kwargs):
            """Gradient function calculation for inner function."""
            variables = grad_kwargs.pop('variables', None)
            if grad_kwargs:
                raise ValueError('Found unexpected kwargs for `grad`: ',
                                 list(grad_kwargs.keys()))
            inputs, seed = list(args), context.seed
            if _in_xla_context():
                inputs = _force_data_dependency(
                    tf.nest.flatten(dresult), inputs + [seed])
                seed = inputs.pop()
            # tf.keras.backend.set_learning_phase(1)
            with tf.GradientTape() as tape:
                tape.watch(inputs)
                if variables is not None:
                    tape.watch(variables)
                with tf.control_dependencies(dresult):
                    with context._replace(is_recomputing=True, seed=seed):
                        result = f(*inputs, **kwargs)
            kw_vars = []
            if variables is not None:
                kw_vars = list(variables)
            grads = tape.gradient(
                result, list(inputs) + kw_vars, output_gradients=dresult)
            return grads[:len(inputs)], grads[len(inputs):]

        return result, grad

    return inner



"""
mkinit geowatch.tasks.fusion --lazy --noattrs -w
mkinit -m geowatch.tasks.fusion --noattrs -w
"""

# Hack to supress pytorch-lightning warning
import setuptools  # NOQA

from geowatch.tasks.fusion import utils
from geowatch.tasks.fusion import datamodules
from geowatch.tasks.fusion import architectures
from geowatch.tasks.fusion import methods


# Not including these removes the runtime warning:
# 'geowatch.tasks.fusion.fit' found in sys.modules after import of package
# 'geowatch.tasks.fusion', but prior to execution of 'geowatch.tasks.fusion.fit';
# this may result in unpredictable behaviour

# from geowatch.tasks.fusion import fit
# from geowatch.tasks.fusion import predict
# from geowatch.tasks.fusion import evaluate

# We can include them via lazy imports. The other packages are not given lazy
# imports so torch.package can introspect them.


def lazy_import(module_name, submodules, submod_attrs, eager='auto'):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    eager_import_flag = False
    if eager == 'auto':
        eager_import_text = os.environ.get('EAGER_IMPORT', '')
        if eager_import_text:
            eager_import_text_ = eager_import_text.lower()
            if eager_import_text_ in {'true', '1', 'on', 'yes'}:
                eager_import_flag = True

        eager_import_module_text = os.environ.get('EAGER_IMPORT_MODULES', '')
        if eager_import_module_text:
            if eager_import_module_text.lower() in __name__.lower():
                eager_import_flag = True
    else:
        eager_import_flag = eager

    if eager_import_flag:
        for name in submodules:
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'architectures',
        'coco_stitcher',
        'datamodules',
        'evaluate',
        'fit',
        'fit_lightning',
        'methods',
        'predict',
        'production',
        'utils',
    },
    submod_attrs={},
)


def __dir__():
    return __all__

__all__ = ['architectures', 'coco_stitcher', 'datamodules', 'evaluate', 'fit',
           'fit_lightning', 'methods', 'predict', 'production', 'utils']



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class DrawRegionCLI(scfg.DataConfig):
    """
    Ignore:
        from geowatch.cli.draw_region import *  # NOQA
        cls = DrawRegionCLI
        cmdline = 0
        kw = kwargs = {}
        kw['models'] = '/home/joncrall/temp/debug_smartflow_latest/ingress/sv_out_region_models/KR_R001.geojson'
        kw['models'] = '/home/joncrall/temp/debug_smartflow_eval20/ingress/sv_out_region_models/KR_R001.geojson'
        kw['models'] = 'sv_out_region_models/KR_R001.geojson'
        kw['models'] = 'sc_out_region_models/KR_R001.geojson'

        kw['models'] = 'cropped_region_models_sc/KR_R001.geojson'
        kw['extra_header'] = '03-cropped_region_models_sc'
        kw['fpath'] = '03-cropped_region_models_sc.png'
        cls.main(cmdline=cmdline, **kwargs)

        kw['models'] = 'sc_out_region_models/KR_R001.geojson'
        kw['extra_header'] = '02-sc_out_region_models'
        kw['fpath'] = '02-sc_out_region_models.png'
        cls.main(cmdline=cmdline, **kwargs)

        kw['models'] = 'sv_out_region_models/KR_R001.geojson'
        kw['extra_header'] = '01-sv_out_region_models'
        kw['fpath'] = '01-sv_out_region_models.png'
        cls.main(cmdline=cmdline, **kwargs)

        ub.cmd('kwimage stack_images *.png')
    """
    # param1 = scfg.Value(None, help='param1')

    models = scfg.Value(None, help='site OR region models coercables (the script will attempt to distinguish them)', nargs='+', position=1)

    site_models = scfg.Value(None, help='site model coercable', nargs='+', alias=['sites'])

    region_models = scfg.Value(None, help='region model coercable', nargs='+', alias=['regions'])
    # viz_dpath = scfg.Value(None, help='if specified will write stats visualizations and plots to this directory')

    extra_header = scfg.Value(None)

    io_workers = scfg.Value('avail', help='number of workers for parallel io')

    with_timeline = scfg.Value(True, help='if True draw the timeline')

    fpath = scfg.Value('auto', help=ub.paragraph(
        '''
        Path to write the visualization image to.
        '''), alias=['output_fpath'])

    sidecar = scfg.Value(False, isflag=True, help=ub.paragraph(
        '''
        if True, then default the output fpath to write viz as a sidecar next
        to the input. When false the default is region.png.
        Has no effect if fpath is specified.
        '''))

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> from geowatch.cli.draw_region import *  # NOQA
            >>> cmdline = 0
            >>> kwargs = dict()
            >>> cls = DrawRegionCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        import rich
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)

        if config.fpath == 'auto':
            if config.sidecar:
                paths = config.models or config.region_models or config.site_models
                if len(paths) > 1:
                    raise Exception('Cannot do an auto sidecar with multiple inputs')
                path = paths[0]
                config.fpath = ub.Path(path).augment(ext='.png')
            else:
                #
                config.fpath = 'region.png'

        rich.print('config = ' + ub.urepr(config, nl=1))

        # import copy
        # import numpy as np
        import pandas as pd
        # from kwutil import util_time
        from geowatch.geoannots import geomodels
        from kwgis.utils import util_gis
        from kwutil import util_parallel
        import kwimage
        import kwplot
        from geowatch.utils import util_kwplot

        # Note: these colors are kinda not useful.
        from geowatch import heuristics
        status_to_color = {r['status']: r['color'] for r in heuristics.HUERISTIC_STATUS_DATA}

        site_models = []
        region_models = []

        io_workers = util_parallel.coerce_num_workers(config['io_workers'])
        print(f'io_workers={io_workers}')

        if config.models:
            if config.site_models:
                raise ValueError('the models and site_models arguments are mutex')
            if config.region_models:
                raise ValueError('the models and region_models arguments are mutex')
            models = list(util_gis.coerce_geojson_datas(config.models, format='json', workers=io_workers))
            for model_info in models:
                model_data = model_info['data']
                model = geomodels.coerce_site_or_region_model(model_data)
                if isinstance(model, geomodels.SiteModel):
                    site_models.append(model)
                elif isinstance(model, geomodels.RegionModel):
                    region_models.append(model)
                else:
                    raise AssertionError
        else:
            site_models = list(geomodels.SiteModel.coerce_multiple(config['site_models'], workers=io_workers))
            region_models = list(geomodels.RegionModel.coerce_multiple(config['region_models'], workers=io_workers))

        dataframes = ub.ddict(list)

        for region in region_models:
            dataframes['region'] += [region.pandas_region()]
            dataframes['site_summary'] += [region.pandas_summaries()]

        for site in site_models:
            dataframes['sites'] += [site.pandas_site()]
            # dataframes['observations'] += [site.pandas_observations()]

        # if dataframes['region']:
        if len(dataframes['region']):
            _region_df = pd.concat(dataframes['region'])
            _summary_df = pd.concat(dataframes['site_summary'], axis=0)
            region_df = util_gis.project_gdf_to_local_utm(_region_df)
            summary_df = _summary_df.to_crs(region_df.crs)
        else:
            region_df = None
            summary_df = None

        if len(dataframes['sites']):
            _site_df = pd.concat(dataframes['sites'])
            site_df = util_gis.project_gdf_to_local_utm(_site_df)
        else:
            site_df = None

        # if config.viz_dpath is None:
        #     config.viz_dpath = ub.Path('.').resolve()

        region_ids = set()
        if region_df is not None:
            region_ids.update(set(region_df['region_id']))
        if site_df is not None:
            region_ids.update(set(site_df['region_id']))

        unique_status = set()
        if summary_df is not None:
            unique_status.update(summary_df['status'].unique())
        if site_df is not None:
            unique_status.update(site_df['status'].unique())

        draw_backend = 'cv2'
        draw_backend = 'mpl'
        if draw_backend == 'mpl':
            kwplot.autosns()

            title = util_kwplot.TitleBuilder()

            if len(str(config.models)) < 255:
                title.ensure_newline()
                title.add_part(str(config.models))
                title.ensure_newline()

            region_ids = sorted(set(region_ids))
            if len(region_ids) == 1:
                region_title = f'Region: {region_ids[0]}'
            else:
                region_title = f'Regions: {",".join(region_ids)}'
            title.add_part(region_title)

            if config.extra_header is not None:
                title.add_part(config.extra_header)

            extra_header = ub.Path('.').resolve().parent.name
            title.add_part(extra_header)
            # extra_header = 'Latest'
            # title_row.append(f'{geowatch.__version__=}')

            unique_colors = [kwimage.Color.coerce(status_to_color.get(s, 'cyan')) for s in unique_status]
            unique_status_to_color = ub.dzip(unique_status, unique_colors)
            legend_canvas = kwplot.make_legend_img(unique_status_to_color, dpi=600)

            # Assign a unique color to each site (Used for edges)
            unique_site_ids = set()
            if site_df is not None:
                unique_site_ids.update(site_df.site_id)
            if summary_df is not None:
                unique_site_ids.update(summary_df.site_id)
            _unique_colors = kwimage.Color.distinct(len(unique_site_ids))
            # Darken the edge colors so they arent so distracting
            _unique_colors = [kwimage.Color(c).adjust(lighten=-0.25).as01() for c in _unique_colors]
            siteid_to_color = ub.dzip(unique_site_ids, _unique_colors)

            if summary_df is not None:
                summary_df['edge_color'] = summary_df['site_id'].apply(siteid_to_color.__getitem__)
            if site_df is not None:
                site_df['edge_color'] = site_df['site_id'].apply(siteid_to_color.__getitem__)

            # Assign face colors to each site summary
            unique_status_to_face_color01 = {s: c.as01() for s, c in unique_status_to_color.items()}
            # unique_status_to_edge_color01 = {s: c.adjust(lighten=-.05).as01() for s, c in unique_status_to_color.items()}
            if summary_df is not None:
                summary_df['face_color'] = summary_df['status'].apply(unique_status_to_face_color01.__getitem__)
                # summary_df['edge_color'] = summary_df['status'].apply(unique_status_to_edge_color01.__getitem__)
            if site_df is not None:
                site_df['face_color'] = site_df['status'].apply(unique_status_to_face_color01.__getitem__)
                # site_df['edge_color'] = site_df['status'].apply(unique_status_to_edge_color01.__getitem__)

            num_rows = 2 if config.with_timeline else 1
            figman = util_kwplot.FigureManager(
                dpath=ub.Path('.'),
                dpi=300,
                size_inches=(12, 8),
                verbose=True,
            )
            fig = figman.figure(fnum=1, pnum=(num_rows, 2, 1), doclf=True)
            ax = fig.gca()

            # Plot Region Bounds
            if region_df is not None:
                region_df['geometry'].plot(
                    facecolor='none',
                    edgecolor='black',
                    ax=ax
                )

            if summary_df is not None:
                # Plot Site Summary Bounds
                summary_df['geometry'].plot(
                    facecolor=summary_df['face_color'],
                    edgecolor=summary_df['edge_color'],
                    # linewidth=1,
                    alpha=0.5,
                    ax=ax
                )
            if site_df is not None:
                # Plot Site Bounds
                # Note: if both sites and sites summary exist, this will
                # overlay both, which might not look very nice.
                site_df['geometry'].plot(
                    facecolor=site_df['face_color'],
                    edgecolor=site_df['edge_color'],
                    # linewidth=1,
                    alpha=0.5,
                    ax=ax
                )

            miny, maxy = ax.get_ylim()
            ax = figman.figure(fnum=1, pnum=(num_rows, 2, 2)).gca()
            kwplot.imshow(legend_canvas, ax=ax)

            figman.set_figtitle(str(title))

            if config.with_timeline:
                ax = kwplot.figure(fnum=1, pnum=(num_rows, 1, 2), docla=1).gca()

                if site_df is not None:
                    # centroid = site_df.geometry.centroid
                    # site_df['y'] = centroid.y
                    # sub = site_df[['site_id', 'y']]
                    # stacked = pd.concat([sub] * 2, ignore_index=True)
                    # stacked['date'] = pd.concat([site_df.start_date, site_df.end_date], ignore_index=True)
                    # stacked['frame_idx'] = np.concatenate([[0] * len(site_df), [1] * len(site_df)])
                    # sns.lineplot(data=stacked, x='date', y='y', ax=ax, hue='site_id', legend=False)
                    lines = []
                    for idx, row in enumerate(site_df.to_dict('records')):
                        y = row['geometry'].centroid.y
                        color = row['edge_color']
                        xs = [row['start_date'], row['end_date']]
                        xs = util_kwplot.fix_matplotlib_dates(xs)
                        line = {
                            'ys': [y, y],
                            'xs': xs,
                            'color': color,
                            'marker': '.'
                        }
                        lines.append(line)
                    _draw_lines(lines, ax)
                    # TODO: make this formatter fixup work better.
                    import matplotlib.dates as mdates
                    ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))

                if summary_df is not None:
                    lines = []
                    for idx, row in enumerate(summary_df.to_dict('records')):
                        y = row['geometry'].centroid.y
                        color = row['edge_color']
                        xs = [row['start_date'], row['end_date']]
                        xs = util_kwplot.fix_matplotlib_dates(xs)
                        line = {
                            'ys': [y, y],
                            'xs': xs,
                            'color': color,
                            'marker': '.'
                        }
                        lines.append(line)
                    _draw_lines(lines, ax)
                    # TODO: make this formatter fixup work better.
                    import matplotlib.dates as mdates
                    ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
                    # ax.xaxis.set_major_locator(mdates.DayLocator(interval=90))

                    # centroid = summary_df.geometry.centroid
                    # summary_df['y'] = centroid.y
                    # sub = summary_df[['site_id', 'y']]
                    # stacked = pd.concat([sub] * 2, ignore_index=True)
                    # stacked['date'] = pd.concat([summary_df.start_date, summary_df.end_date], ignore_index=True)
                    # stacked['frame_idx'] = np.concatenate([[0] * len(summary_df), [1] * len(summary_df)])
                    # sns.lineplot(data=stacked, x='date', y='y', ax=ax, hue='site_id', legend=False)

                ax.set_ylim(miny, maxy)

            fpath = config.fpath
            final_fpath = figman.finalize(fpath)
            dpath = final_fpath.absolute().parent
            rich.print(f'Wrote to: [link={dpath}]{dpath}[/link]')

        else:
            raise NotImplementedError


def _demo_draw_times():
    """
    MWE code to help get timelines to draw right.
    """
    lines = [
        {'ys': [5178041.5354517745, 5178041.5354517745],
         'xs': [16530.0, 18745.0],
         'color': (0.0, 0.38171248398857033, 0.7303921568627451),
         'marker': '.'},
        {'ys': [5177799.260253956, 5177799.260253956],
         'xs': [16050.0, 18901.0],
         'color': (0.0, 0.38171248398857033, 0.7303921568627451),
         'marker': '.'},
        {'ys': [5177300.748245601, 5177300.748245601],
         'xs': [16050.0, 18536.0],
         'color': (0.0, 0.38171248398857033, 0.7303921568627451),
         'marker': '.'},
        {'ys': [5177175.214791525, 5177175.214791525],
         'xs': [16530.0, 18901.0],
         'color': (0.0, 0.38171248398857033, 0.7303921568627451),
         'marker': '.'},
        {'ys': [5176775.194194238, 5176775.194194238],
         'xs': [16050.0, 18745.0],
         'color': (0.0, 0.38171248398857033, 0.7303921568627451),
         'marker': '.'},
        {'ys': [5176569.654561392, 5176569.654561392],
         'xs': [16530.0, 18911.0],
         'color': (0.0, 0.38171248398857033, 0.7303921568627451),
         'marker': '.'},
    ]
    import kwplot
    kwplot.autosns()
    fig = kwplot.figure(doclf=1)
    ax = fig.gca()
    _draw_lines(lines, ax)
    # TODO: make this formatter fixup work better.
    import matplotlib.dates as mdates
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))

    ...


def _draw_lines(lines, ax):
    # TODO: could use util_kwplot artist manager here
    for line in lines:
        xs = line['xs']
        ys = line['ys']
        linestyle = line.get('linestyle', '-')
        ax.plot(xs, ys, linestyle, marker=line['marker'], color=line['color'])


__cli__ = DrawRegionCLI
main = __cli__.main

if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/geowatch/geowatch/cli/draw_region.py
        python -m geowatch.cli.draw_region
    """
    main()



# ---#
#
# Initial code from https://github.com/neptune-ai/open-solution-mapping-challenge/blob/master/src/unet_models.py
#
# ---
from collections import OrderedDict
from torch import nn
from torch.nn import functional as F
import torch
import torchvision


def conv3x3(in_, out):
    return nn.Conv2d(in_, out, 3, padding=1)


class ConvRelu(nn.Module):
    def __init__(self, in_, out):
        super().__init__()
        self.conv = conv3x3(in_, out)
        self.activation = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.activation(x)
        print("ConvRelu:", x.shape, x.min().item(), x.max().item())
        return x


class DecoderBlockV2(nn.Module):
    def __init__(self, in_channels, middle_channels, out_channels, is_deconv=True):
        super(DecoderBlockV2, self).__init__()
        self.in_channels = in_channels

        if is_deconv:
            """
                Paramaters for Deconvolution were chosen to avoid artifacts, following
                link https://distill.pub/2016/deconv-checkerboard/
            """

            self.block = nn.Sequential(
                ConvRelu(in_channels, middle_channels),
                nn.ConvTranspose2d(middle_channels, out_channels, kernel_size=4, stride=2,
                                   padding=1),
                nn.ReLU(inplace=True)
            )
        else:
            self.block = nn.Sequential(
                nn.Upsample(scale_factor=2, mode='bilinear', align_corners=False),
                nn.Conv2d(in_channels, middle_channels, 3, padding=1, bias=True),
                nn.BatchNorm2d(middle_channels),
                nn.ELU(),
                nn.Conv2d(middle_channels, out_channels, 3, padding=1, bias=True),
                nn.BatchNorm2d(out_channels),
                nn.ELU()
            )

    def forward(self, x):
        return self.block(x)


def cat_non_matching(x1, x2):
    diffY = x1.size()[2] - x2.size()[2]
    diffX = x1.size()[3] - x2.size()[3]

    x2 = F.pad(x2, (diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2))

    # for padding issues, see
    # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
    # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd

    x = torch.cat([x1, x2], dim=1)
    return x


class UNetResNetBackbone(nn.Module):
    """PyTorch U-Net model using ResNet(34, 101 or 152) encoder.
    UNet: https://arxiv.org/abs/1505.04597
    ResNet: https://arxiv.org/abs/1512.03385
    Proposed by Alexander Buslaev: https://www.linkedin.com/in/al-buslaev/
    Args:
            encoder_depth (int): Depth of a ResNet encoder (34, 101 or 152).
            num_filters (int, optional): Number of filters in the last layer of decoder. Defaults to 32.
            dropout_2d (float, optional): Probability factor of dropout layer before output layer. Defaults to 0.2.
            pretrained (bool, optional):
                False - no pre-trained weights are being used.
                True  - ResNet encoder is pre-trained on ImageNet.
                Defaults to False.
            is_deconv (bool, optional):
                False: bilinear interpolation is used in decoder.
                True: deconvolution is used in decoder.
                Defaults to False.
    """

    def __init__(self, encoder_depth, num_filters=32, dropout_2d=0.2,
                 pretrained=False, is_deconv=False):
        super().__init__()
        self.dropout_2d = dropout_2d

        if encoder_depth == 34:
            self.encoder = torchvision.models.resnet34(pretrained=pretrained)
            bottom_channel_nr = 512
        elif encoder_depth == 101:
            self.encoder = torchvision.models.resnet101(pretrained=pretrained)
            bottom_channel_nr = 2048
        elif encoder_depth == 152:
            self.encoder = torchvision.models.resnet152(pretrained=pretrained)
            bottom_channel_nr = 2048
        else:
            raise NotImplementedError('only 34, 101, 152 version of Resnet are implemented')

        self.pool = nn.MaxPool2d(2, 2)

        self.relu = nn.ReLU(inplace=True)

        self.conv1 = nn.Sequential(self.encoder.conv1,
                                   self.encoder.bn1,
                                   self.encoder.relu,
                                   self.pool)

        self.conv2 = self.encoder.layer1

        self.conv3 = self.encoder.layer2

        self.conv4 = self.encoder.layer3

        self.conv5 = self.encoder.layer4

        self.center = DecoderBlockV2(bottom_channel_nr, num_filters * 8 * 2, num_filters * 8, is_deconv)
        self.dec5 = DecoderBlockV2(bottom_channel_nr + num_filters * 8, num_filters * 8 * 2, num_filters * 8, is_deconv)
        self.dec4 = DecoderBlockV2(bottom_channel_nr // 2 + num_filters * 8, num_filters * 8 * 2, num_filters * 8,
                                   is_deconv)
        self.dec3 = DecoderBlockV2(bottom_channel_nr // 4 + num_filters * 8, num_filters * 4 * 2, num_filters * 2,
                                   is_deconv)
        self.dec2 = DecoderBlockV2(bottom_channel_nr // 8 + num_filters * 2, num_filters * 2 * 2, num_filters * 2 * 2,
                                   is_deconv)
        self.dec1 = DecoderBlockV2(num_filters * 2 * 2, num_filters * 2 * 2, num_filters, is_deconv)

    def forward(self, x):
        # print("x:", x.shape, x.min().item(), x.max().item())
        conv1 = self.conv1(x)
        # print("conv1:", conv1.shape, conv1.min().item(), conv1.max().item())
        conv2 = self.conv2(conv1)
        # print("conv2:", conv2.shape, conv2.min().item(), conv2.max().item())
        conv3 = self.conv3(conv2)
        # print("conv3:", conv3.shape, conv3.min().item(), conv3.max().item())
        conv4 = self.conv4(conv3)
        # print("conv4:", conv4.shape, conv4.min().item(), conv4.max().item())
        conv5 = self.conv5(conv4)
        # print("conv5:", conv5.shape, conv5.min().item(), conv5.max().item())

        pool = self.pool(conv5)
        # print("pool:", pool.shape, pool.min().item(), pool.max().item())
        center = self.center(pool)
        # print("center:", center.shape, center.min().item(), center.max().item())

        dec5 = self.dec5(cat_non_matching(conv5, center))
        # print("center:", center.shape, center.min().item(), center.max().item())

        dec4 = self.dec4(cat_non_matching(conv4, dec5))
        # print("dec4:", dec4.shape, dec4.min().item(), dec4.max().item())
        dec3 = self.dec3(cat_non_matching(conv3, dec4))
        # print("dec3:", dec3.shape, dec3.min().item(), dec3.max().item())
        dec2 = self.dec2(cat_non_matching(conv2, dec3))
        # print("dec2:", dec2.shape, dec2.min().item(), dec2.max().item())
        dec1 = self.dec1(dec2)
        # print("dec1:", dec1.shape, dec1.min().item(), dec1.max().item())

        y = F.dropout2d(dec1, p=self.dropout_2d)
        # print("y:", y.shape, y.min().item(), y.max().item())

        result = OrderedDict()
        result["out"] = y

        return result
